import asyncio
import base64
import typing as t
from contextlib import suppress

import aiohttp
from yarl import URL

from ..exceptions import TonConnectError
from ..models import IncomingMessage
from ..utils import add_path_to_url
from .source import (
    EventMessage,
    EventSource,
    ReadyState,
)
from .storage import ProviderStorage

_T = t.TypeVar("_T")


async def retry_until_success(
    fn: t.Callable[[], t.Awaitable[_T]],
    *,
    attempts: int = 10,
    delay: float = 0.2,
) -> _T:
    """Retry an async callable until it succeeds.

    :param fn: Async callable to retry.
    :param attempts: Maximum number of attempts.
    :param delay: Delay between attempts in seconds.
    :return: Result of *fn*.
    :raises BaseException: Last exception if all attempts fail.
    """
    last_error: BaseException | None = None

    for i in range(attempts):
        try:
            return await fn()
        except BaseException as exc:  # noqa: PERF203
            last_error = exc
            if i + 1 >= attempts:
                break
            await asyncio.sleep(delay)

    assert last_error is not None
    raise last_error


class Gateway:
    """SSE bridge gateway for a single TonConnect bridge connection."""

    SSE_PATH = "events"
    POST_PATH = "message"
    HEARTBEAT_MESSAGE = "heartbeat"

    RECONNECT_ATTEMPTS: int = 5
    RECONNECT_DELAY: float = 2.0

    def __init__(
        self,
        *,
        session_id: str,
        bridge_url: str,
        storage: ProviderStorage,
        headers: dict[str, str] | None = None,
        on_gateway_message: t.Callable[[IncomingMessage, str], t.Awaitable[None]],
        on_gateway_error: t.Callable[[Exception], t.Awaitable[None]],
    ) -> None:
        """Initialize the gateway.

        :param session_id: SSE client identifier.
        :param bridge_url: Bridge base URL.
        :param storage: Provider storage.
        :param headers: Extra HTTP headers, or ``None``.
        :param on_gateway_message: Async callback for incoming messages.
        :param on_gateway_error: Async callback for errors.
        """
        self._session_id = session_id
        self._bridge_url = bridge_url
        self._storage = storage

        self._headers = headers or {}
        self._on_gateway_message = on_gateway_message
        self._on_gateway_error = on_gateway_error

        self._http_session: aiohttp.ClientSession | None = None
        self._event_source: EventSource | None = None

        self._connect_lock = asyncio.Lock()
        self._reconnect_task: asyncio.Task[None] | None = None

        self._closed = False
        self._paused = False

    @property
    def bridge_url(self) -> str:
        """Bridge URL this gateway is connected to."""
        return self._bridge_url

    @property
    def ready_state(self) -> ReadyState:
        """Current SSE connection state."""
        if self._closed:
            return ReadyState.CLOSED
        es = self._event_source
        if es is None:
            return ReadyState.CLOSED
        return es.ready_state

    async def register_session(self, timeout: float = 12.0) -> None:
        """Open an SSE session with the bridge.

        :param timeout: Connection timeout in seconds.
        :raises TonConnectError: If the gateway is closed.
        """
        if self._closed:
            raise TonConnectError("Bridge is closed")
        if self._paused:
            return

        async with self._connect_lock:
            if self._closed or self._paused:  # race condition protection
                return  # type: ignore[unreachable]

            state = self.ready_state
            if state in (ReadyState.OPEN, ReadyState.CONNECTING):
                return

            await self._close_event_source()

            last_event_id = await self._storage.get_last_event_id()
            url = self._make_open_url(last_event_id)
            session = await self._ensure_session()
            es = EventSource(
                session=session,
                url=url,
                headers=self._headers,
                on_message=self._on_event_source_message,
                on_error=self._on_event_source_error,
            )
            self._event_source = es

            try:
                await es.open(timeout)
            except BaseException:
                await self._close_event_source()
                raise

    async def send(
        self,
        message: bytes,
        receiver: str,
        topic: str,
        *,
        ttl: int = 300,
        attempts: int = 10,
        delay: float = 5.0,
    ) -> None:
        """Send an encrypted message to a wallet via the bridge.

        :param message: Encrypted message bytes.
        :param receiver: Hex-encoded wallet public key.
        :param topic: Bridge topic (RPC method name).
        :param ttl: Message time-to-live in seconds.
        :param attempts: Retry attempts for the HTTP POST.
        :param delay: Delay between retries in seconds.
        :raises TonConnectError: If the gateway is closed or send fails.
        """
        if self._closed:
            raise TonConnectError("Bridge is closed")

        url = self._make_send_url(
            receiver=receiver,
            topic=topic,
            ttl=ttl,
        )
        body = base64.b64encode(message).decode()

        async def _post_once() -> None:
            session = await self._ensure_session()
            async with session.post(str(url), data=body) as resp:
                if resp.status < 200 or resp.status >= 300:
                    raise TonConnectError(f"Bridge send failed, status {resp.status}")

        await retry_until_success(
            _post_once,
            attempts=attempts,
            delay=delay,
        )

    async def pause(self) -> None:
        """Pause SSE listening and cancel any pending reconnect."""
        self._paused = True

        task, self._reconnect_task = self._reconnect_task, None
        if task is not None:
            task.cancel()
            with suppress(asyncio.CancelledError, Exception):
                await task

        await self._close_event_source()

    async def unpause(self) -> None:
        """Resume SSE listening after a pause."""
        if self._closed or not self._paused:
            return
        self._paused = False
        await self.register_session()

    async def close(self) -> None:
        """Permanently close the gateway and release resources."""
        self._closed = True
        self._paused = True

        task, self._reconnect_task = self._reconnect_task, None
        if task is not None:
            task.cancel()
            with suppress(asyncio.CancelledError, Exception):
                await task

        await self._close_event_source()
        await self._close_session()

    async def _reconnect(self) -> None:
        """Schedule automatic reconnection with retries."""
        if self._closed or self._paused:
            return

        task = self._reconnect_task
        if task is not None and not task.done():
            return

        async def _runner() -> None:
            await self._close_event_source()
            await asyncio.sleep(self.RECONNECT_DELAY)
            if self._closed or self._paused:
                return
            try:
                await retry_until_success(
                    self.register_session,
                    attempts=self.RECONNECT_ATTEMPTS,
                    delay=self.RECONNECT_DELAY,
                )
            except Exception as exc:
                if not self._closed and not self._paused:
                    with suppress(Exception):
                        err = TonConnectError(
                            f"Bridge reconnect failed after {self.RECONNECT_ATTEMPTS} attempts: {exc}"
                        )
                        await self._on_gateway_error(err)

        task = asyncio.create_task(_runner())
        self._reconnect_task = task

    async def _ensure_session(self) -> aiohttp.ClientSession:
        """Return the HTTP session, creating one if needed."""
        if self._http_session is None or self._http_session.closed:
            self._http_session = aiohttp.ClientSession(
                headers=self._headers,
                timeout=aiohttp.ClientTimeout(
                    total=None,
                    sock_connect=30,
                    sock_read=None,
                ),
            )
        return self._http_session

    async def _close_event_source(self) -> None:
        """Close and discard the current ``EventSource``."""
        es, self._event_source = self._event_source, None
        if es is None:
            return
        with suppress(Exception):
            await es.close()

    async def _close_session(self) -> None:
        """Close and discard the HTTP session."""
        session, self._http_session = self._http_session, None
        if session is None:
            return
        with suppress(Exception):
            await session.close()

    async def _on_event_source_error(self, exc: Exception) -> None:
        """Handle SSE-level errors, triggering reconnect as needed."""
        if self._closed or self._paused:
            return

        state = self.ready_state
        if state == ReadyState.CONNECTING:
            with suppress(Exception):
                err = TonConnectError("Bridge error, failed to connect")
                await self._on_gateway_error(err)
            await self._reconnect()
            return

        if state == ReadyState.OPEN:
            with suppress(Exception):
                await self._on_gateway_error(exc)
            return

        await self._reconnect()

    async def _on_event_source_message(self, e: EventMessage) -> None:
        """Handle a raw SSE event, parsing and dispatching it."""
        if e.data is None or e.data == self.HEARTBEAT_MESSAGE:
            return
        if e.event_id:
            await self._storage.store_last_event_id(e.event_id)

        if self._closed or self._paused:
            return

        try:
            msg = IncomingMessage.model_validate_json(e.data)
        except Exception as exc:
            with suppress(Exception):
                err = TonConnectError(f"Bridge message parse failed, message: `{e.data}`, error: {exc}")
                await self._on_gateway_error(err)
            return

        try:
            await self._on_gateway_message(msg, self._bridge_url)
        except Exception as exc:
            with suppress(Exception):
                await self._on_gateway_error(exc)

    def _make_open_url(self, last_event_id: str | None = None) -> URL:
        """Build the SSE endpoint URL."""
        query: dict[str, str] = {"client_id": self._session_id}
        if last_event_id:
            query["last_event_id"] = last_event_id
        base = add_path_to_url(self._bridge_url, self.SSE_PATH)
        return URL(base).with_query(query)

    def _make_send_url(self, receiver: str, topic: str, ttl: int) -> URL:
        """Build the POST endpoint URL for sending messages."""
        query: dict[str, str] = {
            "client_id": self._session_id,
            "to": receiver,
            "ttl": str(ttl),
            "topic": topic,
        }
        base = add_path_to_url(self._bridge_url, self.POST_PATH)
        return URL(base).with_query(query)



import typing as t

from pydantic import Field, TypeAdapter, field_validator

from ._types import A, BaseModel
from .feature import FeatureType, FeatureTypes

_FEATURE_ADAPTER: TypeAdapter[FeatureType] = TypeAdapter(FeatureType)


class Device(BaseModel):
    """Wallet device information from the connect event."""

    platform: str
    """Device platform identifier."""
    app_name: str = A("appName")
    """Wallet application name."""
    app_version: str = A("appVersion")
    """Wallet application version."""
    max_protocol_version: int = A("maxProtocolVersion")
    """Maximum supported TonConnect protocol version."""
    features: FeatureTypes = Field(default_factory=list)
    """Declared wallet features."""

    @field_validator("features", mode="before")
    @classmethod
    def _v_features(cls, v: t.Any) -> FeatureTypes:
        return [_FEATURE_ADAPTER.validate_python(f) for f in v if isinstance(f, dict)]



import asyncio
import time
import typing as t
from enum import Enum

from ton_core import NetworkGlobalID

from .exceptions import (
    RequestTimeoutError,
    TonConnectError,
    TonConnectErrors,
    WalletAlreadyConnectedError,
    WalletNotConnectedError,
)
from .models import (
    Account,
    ActiveConnection,
    AppWallet,
    AppWallets,
    ConnectEventError,
    ConnectEventSuccess,
    ConnectItem,
    ConnectRequest,
    DisconnectEventError,
    DisconnectEventSuccess,
    RpcRequestBase,
    RpcResponseSuccessBase,
    SendTransactionPayload,
    SendTransactionResult,
    SendTransactionRpcRequest,
    SignDataPayload,
    SignDataResult,
    SignDataRpcRequest,
    TonAddressItem,
    TonProofItem,
    Wallet,
    WalletMessage,
    WalletResponseError,
    WalletResponseSuccess,
)
from .provider import Provider
from .provider.storage import ProviderStorage
from .utils import (
    STANDARD_UNIVERSAL_LINK,
    generate_universal_link,
    verify_send_transaction_support,
    verify_sign_data_support,
    verify_wallet_features,
    verify_wallet_network,
)

_ConnectResult = tuple[
    Wallet | None,
    TonConnectError | None,
]
_SendTransactionResult = tuple[
    SendTransactionResult | None,
    TonConnectError | None,
]
_SignDataResult = tuple[
    SignDataResult | None,
    TonConnectError | None,
]
_RequestResult = _SendTransactionResult | _SignDataResult


class Event(str, Enum):
    """Events dispatched by ``Connector``."""

    CONNECT = "connect"
    """Wallet connection result."""
    DISCONNECT = "disconnect"
    """Wallet disconnection result."""
    TRANSACTION = "transaction"
    """Send transaction result."""
    SIGN_DATA = "sign_data"
    """Sign data result."""
    MESSAGE = "message"
    """Raw wallet message before processing."""
    ERROR = "error"
    """Provider/bridge-level or unhandled handler error."""


class Connector:
    """TonConnect session connector managing wallet interactions."""

    DEFAULT_CONNECT_TIMEOUT: float = 15 * 60
    DEFAULT_REQUEST_TIMEOUT: float = 5 * 60

    def __init__(
        self,
        storage: ProviderStorage,
        session_key: str,
        manifest_url: str,
        app_wallets: AppWallets,
        handlers: dict[Event, t.Callable[..., t.Awaitable[None]]],
        headers: dict[str, str] | None = None,
        context: dict[str, t.Any] | None = None,
    ) -> None:
        """Initialize the connector.

        :param storage: Provider storage.
        :param session_key: Unique session key.
        :param manifest_url: URL to ``tonconnect-manifest.json``.
        :param app_wallets: Wallet descriptors passed to the provider as bridge connection sources.
        :param handlers: Event handler mapping.
        :param headers: Extra HTTP headers, or ``None``.
        :param context: User context dict, or ``None``.
        """
        self._session_key = session_key
        self._manifest_url = manifest_url
        self._app_wallets = app_wallets
        self._handlers = handlers
        self._context: dict[str, t.Any] = context or {}

        self._wallet: Wallet | None = None
        self._network: NetworkGlobalID | None = None

        self._connect_future: asyncio.Future[_ConnectResult] | None = None
        self._connect_timeout_task: asyncio.Task[None] | None = None

        self._request_futures: dict[int, asyncio.Future[_RequestResult]] = {}
        self._request_timeout_tasks: dict[int, asyncio.Task[None]] = {}
        self._request_events: dict[int, Event] = {}

        self._storage = storage
        self._provider = Provider(
            storage=self._storage,
            app_wallets=app_wallets,
            on_message=self._on_message,
            on_error=self._on_provider_error,
            headers=headers,
        )

    def __getitem__(self, key: str) -> t.Any:
        """Get a context value by key."""
        return self._context[key]

    def __setitem__(self, key: str, value: t.Any) -> None:
        """Set a context value by key."""
        self._context[key] = value

    @property
    def session_key(self) -> str:
        """Unique session key."""
        return self._session_key

    @property
    def connected(self) -> bool:
        """Whether a wallet is currently connected."""
        return self._wallet is not None

    @property
    def storage(self) -> ProviderStorage:
        """Provider storage instance."""
        return self._storage

    @property
    def wallet(self) -> Wallet | None:
        """Connected wallet, or ``None``."""
        return self._wallet

    @property
    def account(self) -> Account | None:
        """Connected wallet account, or ``None``."""
        return self._wallet.account if self._wallet else None

    @property
    def app_wallet(self) -> AppWallet | None:
        """Matched wallet app descriptor, or ``None``."""
        bridge_url = self._provider.bridge_url
        if bridge_url is None:
            return None
        for w in self._app_wallets:
            if w.bridge_url == bridge_url:
                return w
        return None

    def make_connect_url(
        self,
        request: ConnectRequest,
        app_wallet: AppWallet | None = None,
        redirect_url: str | None = None,
    ) -> str:
        """Generate a universal link for a connect request.

        :param request: Connect request.
        :param app_wallet: Wallet descriptor for custom universal link, or ``None``.
        :param redirect_url: Post-connect redirect URL, or ``None``.
        :return: Link for connecting a wallet.
        """
        universal_link = STANDARD_UNIVERSAL_LINK
        if app_wallet is not None and app_wallet.universal_url:
            universal_link = app_wallet.universal_url

        return generate_universal_link(
            universal_link,
            message=request,
            session_id=self._provider.session_id,
            redirect_url=redirect_url,
        )

    def make_connect_request(
        self,
        ton_proof_payload: str | None = None,
    ) -> ConnectRequest:
        """Build a ``ConnectRequest`` with standard items.

        :param ton_proof_payload: TON Proof challenge payload, or ``None``.
        :return: Connect request.
        """
        items: list[ConnectItem] = [TonAddressItem()]
        if ton_proof_payload is not None:
            items.append(TonProofItem(payload=ton_proof_payload))
        return ConnectRequest(
            manifest_url=self._manifest_url,
            items=items,
        )

    async def connect(
        self,
        request: ConnectRequest,
        network: NetworkGlobalID | None = None,
        app_wallet: AppWallet | None = None,
        redirect_url: str | None = None,
        timeout: float | None = DEFAULT_CONNECT_TIMEOUT,
    ) -> str:
        """Initiate a wallet connection.

        :param request: Connect request.
        :param network: Expected network, or ``None``.
        :param app_wallet: Wallet descriptor for custom universal link, or ``None``.
        :param redirect_url: Post-connect redirect URL, or ``None``.
        :param timeout: Connection timeout in seconds, or ``None``.
        :return: Universal link for connecting a wallet.
        :raises WalletAlreadyConnectedError: If already connected.
        """
        if self.connected:
            raise WalletAlreadyConnectedError()
        self._cancel_connect()
        await self._provider.close_connection()
        await self._provider.connect(request)

        self._network = network
        loop = asyncio.get_running_loop()
        self._connect_future = loop.create_future()

        if timeout is not None:
            self._connect_timeout_task = asyncio.create_task(
                self._run_connect_timeout(timeout),
            )

        return self.make_connect_url(request, app_wallet, redirect_url=redirect_url)

    async def restore(self) -> bool:
        """Restore a connection from storage.

        :return: ``True`` if an active connection was restored.
        """
        connect_event = await self._provider.restore_connection()
        if connect_event is None:
            return False

        wallet = Wallet.from_payload(connect_event.payload)
        self._network = wallet.account.network
        self._wallet = wallet
        return True

    async def disconnect(self) -> None:
        """Disconnect the wallet.

        :raises WalletNotConnectedError: If no wallet is connected.
        """
        if not self.connected:
            raise WalletNotConnectedError()

        await self._provider.disconnect()
        self._wallet = None
        self._network = None
        self._cancel_all_requests()

    async def send_transaction(
        self,
        payload: SendTransactionPayload,
        timeout: float | None = None,
    ) -> int:
        """Request the wallet to sign and send a transaction.

        :param payload: Transaction payload.
        :param timeout: Request timeout in seconds, or ``None``.
        :return: Request ID for ``wait_transaction``.
        """
        has_extra = any(m.extra_currency is not None for m in payload.messages)

        if self._wallet is not None:
            verify_send_transaction_support(
                self._wallet,
                len(payload.messages),
                app_wallet=self.app_wallet,
                has_extra_currency=has_extra,
            )

        if self._network is not None and payload.network is None:
            payload = payload.model_copy(update={"network": self._network})
        if self.account is not None and payload.from_address is None:
            payload = payload.model_copy(update={"from_address": self.account.address})

        if timeout is None and payload.valid_until is not None:
            timeout = max(payload.valid_until - int(time.time()), 0)
        if timeout is None:
            timeout = self.DEFAULT_REQUEST_TIMEOUT  # type: ignore[unreachable]

        return await self._send_request(
            SendTransactionRpcRequest(params=[payload]),
            Event.TRANSACTION,
            timeout,
        )

    async def sign_data(
        self,
        payload: SignDataPayload,
        timeout: float = DEFAULT_REQUEST_TIMEOUT,
    ) -> int:
        """Request the wallet to sign arbitrary data.

        :param payload: Sign data payload.
        :param timeout: Request timeout in seconds, or ``None``.
        :return: Request ID for ``wait_sign_data``.
        """
        if self._wallet is not None:
            verify_sign_data_support(
                self._wallet,
                payload.type,
                app_wallet=self.app_wallet,
            )

        if self._network is not None and payload.network is None:
            payload = payload.model_copy(update={"network": self._network})
        if self.account is not None and payload.from_address is None:
            payload = payload.model_copy(update={"from_address": self.account.address})

        return await self._send_request(
            SignDataRpcRequest(params=[payload]),
            Event.SIGN_DATA,
            timeout,
        )

    async def wait_connect(self) -> _ConnectResult:
        """Await the pending connect result.

        :return: `(wallet, None)` on success, `(None, error)` on failure.
        :raises TonConnectError: If no connect request is pending.
        """
        if self._connect_future is None:
            raise TonConnectError("No pending connect request")
        try:
            return await self._connect_future
        finally:
            self._connect_future = None

    async def wait_transaction(self, request_id: int) -> _SendTransactionResult:
        """Await a pending send transaction result.

        :param request_id: Request ID from ``send_transaction``.
        :return: `(result, None)` on success, `(None, error)` on failure.
        :raises TonConnectError: If request ID is unknown or not a transaction.
        """
        event = self._request_events.get(request_id)
        if event is not Event.TRANSACTION:
            raise TonConnectError(f"Request {request_id} is not a pending transaction")
        return await self._wait_request(request_id)  # type: ignore[return-value]

    async def wait_sign_data(self, request_id: int) -> _SignDataResult:
        """Await a pending sign data result.

        :param request_id: Request ID from ``sign_data``.
        :return: `(result, None)` on success, `(None, error)` on failure.
        :raises TonConnectError: If request ID is unknown or not sign_data.
        """
        event = self._request_events.get(request_id)
        if event is not Event.SIGN_DATA:
            raise TonConnectError(f"Request {request_id} is not a pending sign_data")
        return await self._wait_request(request_id)  # type: ignore[return-value]

    def drop_connect(self) -> None:
        """Cancel the pending connect request."""
        future = self._connect_future
        if future is None:
            return
        self._connect_future = None
        self._cancel_connect_timeout()
        if not future.done():
            future.cancel()

    def drop_request(self, request_id: int) -> None:
        """Cancel a pending request by ID.

        :param request_id: Request ID to cancel.
        """
        self._cancel_request_timeout(request_id)
        self._request_events.pop(request_id, None)
        future = self._request_futures.pop(request_id, None)
        if future is not None and not future.done():
            future.cancel()

    async def close(self) -> None:
        """Close all bridge connections without notifying the wallet."""
        self._cancel_connect()
        self._cancel_all_requests()
        await self._provider.close_connection()

    async def pause(self) -> None:
        """Pause SSE listening."""
        await self._provider.pause()

    async def unpause(self) -> None:
        """Resume SSE listening."""
        await self._provider.unpause()

    async def to_connection(self) -> ActiveConnection | None:
        """Return the active connection from storage, or ``None``."""
        try:
            conn = await self._storage.get_connection()
        except TonConnectError:
            return None
        return conn if isinstance(conn, ActiveConnection) else None

    async def _wait_request(self, request_id: int) -> _RequestResult:
        """Await a pending request result."""
        future = self._request_futures.get(request_id)
        if future is None:
            raise TonConnectError(f"Unknown request id: {request_id}")
        try:
            return await future
        finally:
            if request_id in self._request_futures:
                del self._request_futures[request_id]

    async def _emit_connect(self, error: TonConnectError | None) -> None:
        """Resolve the connect future and dispatch the event."""
        self._cancel_connect_timeout()
        if self._connect_future is not None and not self._connect_future.done():
            self._connect_future.set_result((self._wallet, error))
        await self._dispatch(Event.CONNECT, error)

    async def _emit_request(
        self,
        request_id: int,
        result: t.Any,
        error: TonConnectError | None,
    ) -> None:
        """Resolve a request future and dispatch the event."""
        self._cancel_request_timeout(request_id)
        event = self._request_events.pop(request_id, None)
        future = self._request_futures.pop(request_id, None)
        if future is not None and not future.done():
            future.set_result((result, error))
        if event is not None:
            await self._dispatch(event, request_id, result, error)

    async def _dispatch(self, event: Event, *args: t.Any) -> None:
        """Dispatch an event to its registered handler."""
        handler = self._handlers.get(event)
        if handler is None:
            return
        try:
            await handler(self, *args)
        except Exception as error:
            if event is not Event.ERROR:
                if not isinstance(error, TonConnectError):
                    error = TonConnectError(str(error))
                await self._dispatch(Event.ERROR, error)

    async def _send_request(
        self,
        request: RpcRequestBase,
        event: Event,
        timeout: float | None,
    ) -> int:
        """Encrypt, send, and register a pending request."""
        if not self.connected:
            raise WalletNotConnectedError()

        request_id = await self._provider.request(request)
        loop = asyncio.get_running_loop()

        self._request_futures[request_id] = loop.create_future()
        self._request_events[request_id] = event

        if timeout is not None:
            self._request_timeout_tasks[request_id] = asyncio.create_task(
                self._run_request_timeout(request_id, timeout),
            )

        return request_id

    async def _run_connect_timeout(self, timeout: float) -> None:
        """Fire connect timeout after delay."""
        try:
            await asyncio.sleep(timeout)
        except asyncio.CancelledError:
            return
        self._connect_timeout_task = None
        await self._provider.close_connection()
        await self._emit_connect(RequestTimeoutError())

    async def _run_request_timeout(self, request_id: int, timeout: float) -> None:
        """Fire request timeout after delay."""
        try:
            await asyncio.sleep(timeout)
        except asyncio.CancelledError:
            return
        if request_id in self._request_timeout_tasks:
            del self._request_timeout_tasks[request_id]
        await self._emit_request(request_id, None, RequestTimeoutError())

    def _cancel_connect_timeout(self) -> None:
        """Cancel the connect timeout task."""
        if self._connect_timeout_task is not None:
            self._connect_timeout_task.cancel()
            self._connect_timeout_task = None

    def _cancel_connect(self) -> None:
        """Cancel the pending connect future and timeout."""
        self._cancel_connect_timeout()
        if self._connect_future is not None and not self._connect_future.done():
            self._connect_future.cancel()
        self._connect_future = None

    def _cancel_request_timeout(self, request_id: int) -> None:
        """Cancel the timeout task for a request."""
        task = self._request_timeout_tasks.pop(request_id, None)
        if task is not None:
            task.cancel()

    def _cancel_all_requests(self) -> None:
        """Cancel all pending request futures and timeouts."""
        for future in self._request_futures.values():
            if not future.done():
                future.cancel()
        self._request_futures.clear()
        for task in self._request_timeout_tasks.values():
            task.cancel()
        self._request_timeout_tasks.clear()
        self._request_events.clear()

    def resume_connect(self, timeout: float) -> None:
        """Resume waiting for a pending connect after restart.

        :param timeout: Remaining timeout in seconds.
        """
        self._cancel_connect()
        loop = asyncio.get_running_loop()
        self._connect_future = loop.create_future()
        self._connect_timeout_task = asyncio.create_task(
            self._run_connect_timeout(timeout),
        )

    def resume_request(self, request_id: int, event: Event, timeout: float) -> None:
        """Resume waiting for a pending RPC request after restart.

        :param request_id: The original RPC request ID.
        :param event: Event type (``Event.TRANSACTION`` or ``Event.SIGN_DATA``).
        :param timeout: Remaining timeout in seconds.
        """
        loop = asyncio.get_running_loop()
        self._request_futures[request_id] = loop.create_future()
        self._request_events[request_id] = event
        self._request_timeout_tasks[request_id] = asyncio.create_task(
            self._run_request_timeout(request_id, timeout),
        )

    async def _on_provider_error(self, error: Exception) -> None:
        """Handle provider-level errors."""
        if not isinstance(error, TonConnectError):
            error = TonConnectError(str(error))
        await self._dispatch(Event.ERROR, error)

    async def _on_message(self, message: WalletMessage) -> None:
        """Route an incoming wallet message to the appropriate handler."""
        await self._dispatch(Event.MESSAGE, message)

        if isinstance(message, ConnectEventSuccess):
            await self._handle_connect_success(message)
        elif isinstance(message, ConnectEventError):
            await self._handle_connect_error(message)
        elif isinstance(message, DisconnectEventSuccess):
            await self._handle_disconnect_success(message)
        elif isinstance(message, DisconnectEventError):
            await self._handle_disconnect_error(message)
        elif isinstance(message, RpcResponseSuccessBase):
            await self._handle_response_success(message)
        elif isinstance(message, WalletResponseError):
            await self._handle_response_error(message)
        else:
            error = TonConnectError(f"Unsupported message: {message}")  # type: ignore[unreachable]
            await self._dispatch(Event.ERROR, error)

    async def _handle_connect_success(self, message: ConnectEventSuccess) -> None:
        """Process a successful connect event."""
        wallet = Wallet.from_payload(message.payload)
        try:
            verify_wallet_features(wallet)
            if self._network is not None:
                verify_wallet_network(wallet, self._network)
        except TonConnectError as err:
            await self._storage.remove_connection()
            await self._provider.close_connection()
            await self._emit_connect(err)
            return

        self._wallet = wallet
        await self._emit_connect(None)

    async def _handle_connect_error(self, message: ConnectEventError) -> None:
        """Process a failed connect event."""
        error = TonConnectErrors.from_code(
            message.payload.code,
            message.payload.message,
        )
        await self._emit_connect(error)

    async def _handle_disconnect_success(self, _: DisconnectEventSuccess) -> None:
        """Process a successful disconnect event."""
        self._wallet = None
        self._cancel_all_requests()
        await self._storage.remove_connection()
        await self._provider.close_connection()
        self._cancel_connect_timeout()
        await self._dispatch(Event.DISCONNECT, None)

    async def _handle_disconnect_error(self, message: DisconnectEventError) -> None:
        """Process a failed disconnect event."""
        error = TonConnectErrors.from_code(
            message.payload.code,
            message.payload.message,
        )
        await self._dispatch(Event.DISCONNECT, error)

    async def _handle_response_success(self, message: WalletResponseSuccess) -> None:
        """Process a successful RPC response."""
        await self._emit_request(int(message.id), message.result, None)

    async def _handle_response_error(self, message: WalletResponseError) -> None:
        """Process a failed RPC response."""
        error = TonConnectErrors.from_code(
            message.error.code,
            message.error.message,
        )
        await self._emit_request(int(message.id), None, error)



from ._types import (
    A,
    BaseModel,
    Binary64,
    ChainId,
    TonAddress,
    TonPublicKey,
    WalletStateInit,
)
from .payload import SignDataPayload
from .proof import TonProofData


class SignDataPayloadDto(BaseModel):
    """Verified ``signData`` payload DTO for signature verification."""

    address: TonAddress
    """Wallet address."""
    network: ChainId
    """Network identifier."""
    public_key: TonPublicKey = A("publicKey")
    """Wallet public key."""
    wallet_state_init: WalletStateInit = A("walletStateInit")
    """Wallet ``StateInit``."""
    signature: Binary64
    """Ed25519 signature (64 bytes)."""
    timestamp: int
    """Signing unix timestamp."""
    domain: str
    """dApp domain."""
    payload: SignDataPayload
    """Signed data payload."""


class TonProofPayloadDto(BaseModel):
    """Verified ``ton_proof`` payload DTO for signature verification."""

    address: TonAddress
    """Wallet address."""
    network: ChainId
    """Network identifier."""
    public_key: TonPublicKey = A("publicKey")
    """Wallet public key."""
    wallet_state_init: WalletStateInit = A("walletStateInit")
    """Wallet ``StateInit``."""
    proof: TonProofData
    """TON Proof data."""



import typing as t

from pydantic import Field, TypeAdapter, field_serializer, field_validator

from ._types import A, BaseModel, OptionalChainId
from .payload import SendTransactionPayload, SignDataPayload

TParam = t.TypeVar("TParam")


class RpcRequestBase(BaseModel):
    """Base for all TonConnect RPC requests."""

    id: str | None = None
    """Request identifier, or ``None``."""
    method: str = ""
    """RPC method name."""

    def to_bytes(self, **kwargs: t.Any) -> bytes:
        """Serialize the request to UTF-8 JSON bytes."""
        return self.dump_json(**kwargs).encode()


class RpcRequestWithParams(RpcRequestBase, t.Generic[TParam]):
    """RPC request carrying a typed ``params`` list."""

    _params_adapter: t.ClassVar[TypeAdapter[t.Any]]

    params: list[TParam]
    """Typed parameter list."""

    @classmethod
    def _dump_param(cls, param: TParam) -> str:
        """Serialize a single param to a JSON string."""
        raw = cls._params_adapter.dump_json(param, by_alias=True, exclude_none=True)
        return raw.decode()

    @classmethod
    def _load_param(cls, raw: t.Any) -> TParam:
        """Deserialize a single param from JSON or dict."""
        if isinstance(raw, str):
            return t.cast("TParam", cls._params_adapter.validate_json(raw))
        return t.cast("TParam", cls._params_adapter.validate_python(raw))

    @field_validator("params", mode="before")
    @classmethod
    def _v_params(cls, v: t.Any) -> list[TParam]:
        if v is None:
            return []
        if not isinstance(v, list):
            raise TypeError("params must be a list")
        return [cls._load_param(item) for item in v]

    @field_serializer("params")
    def _s_params(self, v: list[TParam]) -> list[str]:
        return [self._dump_param(item) for item in v]


class TonAddressItem(BaseModel):
    """Connect request item for TON address."""

    name: t.Literal["ton_addr"] = "ton_addr"
    """Item type literal."""
    network: OptionalChainId = None
    """Target network, or ``None``."""


class TonProofItem(BaseModel):
    """Connect request item for TON Proof."""

    name: t.Literal["ton_proof"] = "ton_proof"
    """Item type literal."""
    payload: str
    """Challenge payload string."""


ConnectItem: t.TypeAlias = t.Annotated[
    TonAddressItem | TonProofItem,
    Field(discriminator="name"),
]


class ConnectRequest(BaseModel):
    """TonConnect connect request."""

    manifest_url: str = A("manifestUrl")
    """URL to ``tonconnect-manifest.json``."""
    items: list[ConnectItem]
    """Requested connect items."""


class DisconnectRpcRequest(RpcRequestBase):
    """RPC request to disconnect the wallet."""

    method: t.Literal["disconnect"] = "disconnect"
    """RPC method literal."""
    params: list[t.Any] = Field(default_factory=list, min_length=0, max_length=0)
    """Empty parameter list."""


class SendTransactionRpcRequest(RpcRequestWithParams[SendTransactionPayload]):
    """RPC request to send a transaction."""

    _params_adapter = TypeAdapter(SendTransactionPayload)

    method: t.Literal["sendTransaction"] = "sendTransaction"
    """RPC method literal."""
    params: list[SendTransactionPayload]
    """Transaction payloads."""


class SignDataRpcRequest(RpcRequestWithParams[SignDataPayload]):
    """RPC request to sign data."""

    _params_adapter = TypeAdapter(SignDataPayload)

    method: t.Literal["signData"] = "signData"
    """RPC method literal."""
    params: list[SignDataPayload]
    """Sign data payloads."""


RpcRequests: t.TypeAlias = t.Annotated[
    DisconnectRpcRequest | SendTransactionRpcRequest | SignDataRpcRequest,
    Field(discriminator="method"),
]

AppMessage: t.TypeAlias = ConnectRequest | RpcRequests



import argparse

from .__meta__ import __version__


def main() -> None:
    """CLI entry-point."""
    parser = argparse.ArgumentParser(
        prog="ton-connect",
        description="TON Connect CLI.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"ton-connect {__version__}",
    )
    parser.parse_args()


if __name__ == "__main__":
    main()



from __future__ import annotations

import typing as t

from nacl.public import Box, PrivateKey, PublicKey
from nacl.utils import random as random_bytes
from pydantic import field_serializer, field_validator

from ._types import A, BaseModel


class SessionKeyPair(BaseModel):
    """X25519 key pair for TonConnect session encryption."""

    public_key: PublicKey = A("publicKey")
    """X25519 public key."""
    secret_key: PrivateKey = A("secretKey")
    """X25519 private key."""

    _NONCE_LENGTH: t.ClassVar[int] = 24

    @classmethod
    def generate(cls) -> SessionKeyPair:
        """Generate a new random key pair.

        :return: Fresh ``SessionKeyPair``.
        """
        sk = PrivateKey.generate()
        return cls(public_key=sk.public_key, secret_key=sk)

    @property
    def session_id(self) -> str:
        """Hex-encoded public key used as the session identifier."""
        return self.public_key.encode().hex()

    def encrypt(self, message: bytes, receiver_public_key: PublicKey) -> bytes:
        """Encrypt a message for a receiver.

        :param message: Plaintext bytes.
        :param receiver_public_key: Receiver X25519 public key.
        :return: Nonce-prefixed ciphertext.
        """
        nonce = random_bytes(self._NONCE_LENGTH)
        box = Box(self.secret_key, receiver_public_key)
        ciphertext = box.encrypt(message, nonce).ciphertext
        return nonce + ciphertext

    def decrypt(self, message: bytes, sender_public_key: PublicKey) -> bytes:
        """Decrypt a message from a sender.

        :param message: Nonce-prefixed ciphertext.
        :param sender_public_key: Sender X25519 public key.
        :return: Decrypted plaintext bytes.
        """
        nonce = message[: self._NONCE_LENGTH]
        ciphertext = message[self._NONCE_LENGTH :]
        box = Box(self.secret_key, sender_public_key)
        return box.decrypt(ciphertext, nonce)

    @field_validator("public_key", mode="before")
    @classmethod
    def _v_public_key(cls, v: t.Any) -> PublicKey:
        if isinstance(v, PublicKey):
            return v
        return PublicKey(bytes.fromhex(v))

    @field_validator("secret_key", mode="before")
    @classmethod
    def _v_secret_key(cls, v: t.Any) -> PrivateKey:
        if isinstance(v, PrivateKey):
            return v
        return PrivateKey(bytes.fromhex(v))

    @field_serializer("public_key")
    def _s_public_key(self, v: PublicKey) -> str:
        return v.encode().hex()

    @field_serializer("secret_key")
    def _s_secret_key(self, v: PrivateKey) -> str:
        return v.encode().hex()


class BridgeProviderSession(BaseModel):
    """Active bridge session with wallet encryption keys."""

    session_keypair: SessionKeyPair = A("sessionKeyPair")
    """Session X25519 key pair."""
    wallet_public_key: PublicKey = A("walletPublicKey")
    """Wallet X25519 public key."""
    bridge_url: str = A("bridgeUrl")
    """Bridge base URL."""

    @property
    def receiver_public_key(self) -> PublicKey:
        """Wallet X25519 public key."""
        return self.wallet_public_key

    @property
    def receiver(self) -> str:
        """Hex-encoded wallet public key."""
        return self.receiver_public_key.encode().hex()

    @field_validator("wallet_public_key", mode="before")
    @classmethod
    def _v_wallet_public_key(cls, v: t.Any) -> PublicKey:
        if isinstance(v, PublicKey):
            return v
        return PublicKey(bytes.fromhex(v))

    @field_serializer("wallet_public_key")
    def _s_wallet_public_key(self, v: PublicKey) -> str:
        return v.encode().hex()



import asyncio
import typing as t

from .protocol import StorageProtocol


class MemoryStorage(StorageProtocol):
    """In-memory key-value storage for TonConnect sessions."""

    def __init__(self) -> None:
        """Initialize empty in-memory storage."""
        self._data: dict[str, t.Any] = {}
        self._lock = asyncio.Lock()

    async def set_item(self, key: str, value: t.Any) -> None:
        """Store a value under *key*."""
        async with self._lock:
            self._data[key] = value

    async def get_item(self, key: str) -> t.Any | None:
        """Return the value for *key*, or ``None``."""
        async with self._lock:
            return self._data.get(key)

    async def remove_item(self, key: str) -> None:
        """Remove *key* if present."""
        async with self._lock:
            self._data.pop(key, None)



import typing as t


class StorageProtocol(t.Protocol):
    """Key-value storage for TonConnect session data."""

    async def set_item(self, key: str, value: t.Any) -> None:
        """Store a value.

        :param key: Storage key.
        :param value: Value to store.
        """

    async def get_item(self, key: str) -> t.Any | None:
        """Retrieve a value, or ``None`` if missing.

        :param key: Storage key.
        """

    async def remove_item(self, key: str) -> None:
        """Remove a value.

        :param key: Storage key.
        """



# Copyright (c) 2026 Shon Ness
#
# This source code is licensed under the MIT License found in the
# LICENSE file in the root directory of this source tree.

__version__ = "0.0.2"
__author__ = "nessshon"
__url__ = "https://github.com/nessshon/ton-connect"



import asyncio
import typing as t
from contextlib import suppress

import aiohttp
from yarl import URL

from .decoder import EventDecoder
from .message import EventMessage
from .state import ReadyState


class EventSource:
    """SSE (Server-Sent Events) client for TonConnect bridge."""

    def __init__(
        self,
        url: URL,
        session: aiohttp.ClientSession,
        headers: dict[str, str] | None = None,
        *,
        on_message: t.Callable[[EventMessage], t.Awaitable[None]],
        on_error: t.Callable[[Exception], t.Awaitable[None]],
    ) -> None:
        """Initialize the event source.

        :param url: SSE endpoint URL.
        :param session: HTTP session.
        :param headers: Extra request headers, or ``None``.
        :param on_message: Async callback for received events.
        :param on_error: Async callback for errors.
        """
        self._url = url
        self._session = session
        self._headers = headers or {}

        self._on_message = on_message
        self._on_error = on_error

        self._decoder = EventDecoder()

        self._lock = asyncio.Lock()
        self._state = ReadyState.CLOSED

        self._resp: aiohttp.ClientResponse | None = None
        self._task: asyncio.Task[None] | None = None

    @property
    def ready_state(self) -> ReadyState:
        """Current connection state."""
        return self._state

    async def open(self, timeout: float | None = None) -> None:
        """Open the SSE connection.

        :param timeout: Connection timeout in seconds, or ``None``.
        """
        async with self._lock:
            if self._state != ReadyState.CLOSED:
                return
            self._state = ReadyState.CONNECTING

        resp: aiohttp.ClientResponse | None = None
        try:
            req = self._session.get(
                self._url,
                headers={
                    "Accept": "text/event-stream",
                    "Cache-Control": "no-cache",
                    "Connection": "keep-alive",
                    **self._headers,
                },
            )
            resp = await asyncio.wait_for(req, timeout=timeout) if timeout is not None else await req
            assert resp is not None

            resp.raise_for_status()

            async with self._lock:
                if self._state == ReadyState.CLOSED:
                    return

                self._resp = resp
                self._decoder.reset()
                self._state = ReadyState.OPEN
                self._task = asyncio.create_task(self._loop())
                resp = None
        except asyncio.CancelledError:
            async with self._lock:
                self._state = ReadyState.CLOSED
            raise
        finally:
            if resp is not None:
                with suppress(Exception):
                    resp.close()

    async def close(self) -> None:
        """Close the SSE connection and cancel the read loop."""
        async with self._lock:
            if self._state == ReadyState.CLOSED:
                return

            self._state = ReadyState.CLOSED

            task, self._task = self._task, None
            resp, self._resp = self._resp, None
            self._decoder.reset()

        if task is not None and task is not asyncio.current_task():
            task.cancel()
            with suppress(asyncio.CancelledError):
                await task

        if resp is not None:
            with suppress(Exception):
                resp.close()

    async def _loop(self) -> None:
        """Read SSE stream, dispatch events, and handle errors."""
        exc: Exception | None = None
        try:
            resp = self._resp
            if resp is None:
                return

            async for data in resp.content.iter_any():
                if self._state == ReadyState.CLOSED:
                    return

                for msg in self._decoder.feed(data):
                    try:
                        await self._on_message(msg)
                    except Exception as e:  # noqa: PERF203
                        with suppress(Exception):
                            await self._on_error(e)

            for msg in self._decoder.flush():
                try:
                    await self._on_message(msg)
                except Exception as e:  # noqa: PERF203
                    with suppress(Exception):
                        await self._on_error(e)

            exc = ConnectionError("EventSource connection closed")
        except asyncio.CancelledError:
            return
        except Exception as e:
            exc = e
        finally:
            if exc is not None:
                with suppress(Exception):
                    await self._on_error(exc)
            await self.close()



from ._types import A, BaseModel, Binary64


class TonProofDomain(BaseModel):
    """Domain component of a TON Proof."""

    length_bytes: int = A("lengthBytes")
    """Domain string length in bytes."""
    value: str
    """Domain string."""


class TonProofData(BaseModel):
    """TON Proof data from the wallet."""

    timestamp: int
    """Proof unix timestamp."""
    domain: TonProofDomain
    """Signed domain."""
    signature: Binary64
    """Ed25519 signature (64 bytes)."""
    payload: str
    """Challenge payload string."""



import base64
import typing as t

from nacl.public import PublicKey
from pydantic import field_serializer, field_validator

from ._types import A, BaseModel


class IncomingMessage(BaseModel):
    """Encrypted message received from the TonConnect bridge."""

    from_wallet_public_key: PublicKey = A("from")
    """Sender X25519 public key."""
    message: bytes
    """Encrypted message bytes."""

    @property
    def sender_public_key(self) -> PublicKey:
        """Sender X25519 public key."""
        return self.from_wallet_public_key

    @property
    def sender(self) -> str:
        """Hex-encoded sender public key."""
        return self.sender_public_key.encode().hex()

    @field_validator("from_wallet_public_key", mode="before")
    @classmethod
    def _v_from_wallet_public_key(cls, v: t.Any) -> PublicKey:
        if isinstance(v, PublicKey):
            return v
        return PublicKey(bytes.fromhex(v))

    @field_validator("message", mode="before")
    @classmethod
    def _v_message(cls, v: t.Any) -> bytes:
        if isinstance(v, bytes):
            return v
        return base64.b64decode(v, validate=True)

    @field_serializer("from_wallet_public_key")
    def _s_public_key(self, v: PublicKey) -> str:
        return v.encode().hex()

    @field_serializer("message")
    def _s_message(self, v: bytes) -> str:
        return base64.b64encode(v).decode("ascii")



import typing as t

from pydantic import Field

from ._types import A, BaseModel
from .feature import FeatureTypes


class JSBridgeType(BaseModel):
    """JS-bridge connection descriptor."""

    type: t.Literal["js"]
    """Bridge type literal."""
    key: str
    """JS-bridge injection key."""


class SSEBridgeType(BaseModel):
    """SSE-bridge connection descriptor."""

    type: t.Literal["sse"]
    """Bridge type literal."""
    url: str
    """SSE bridge URL."""


BridgeType: t.TypeAlias = t.Annotated[
    JSBridgeType | SSEBridgeType,
    Field(discriminator="type"),
]
BridgeTypes = list[BridgeType]


class AppWallet(BaseModel):
    """Wallet application descriptor from the wallets catalogue."""

    name: str
    """Display name."""
    image: str
    """Icon URL."""
    app_name: str
    """Machine-readable application name."""
    bridge: BridgeTypes
    """Supported bridge types."""
    tondns: str | None = None
    """TON DNS name, or ``None``."""
    about_url: str | None = None
    """About page URL, or ``None``."""
    universal_url: str | None = None
    """Universal link base, or ``None``."""
    deep_link: str | None = A("deepLink", default=None)
    """Deep link scheme, or ``None``."""
    platforms: list[str]
    """Supported platform identifiers."""
    features: FeatureTypes
    """Declared wallet features."""

    @property
    def bridge_url(self) -> str | None:
        """First SSE bridge URL, or ``None``."""
        for b in self.bridge:
            if b.type == "sse":
                return b.url
        return None


AppWallets = list[AppWallet]



from urllib.parse import (
    parse_qsl,
    urlencode,
    urlparse,
    urlsplit,
    urlunsplit,
)

from ..models.request import ConnectRequest

TONCONNECT_PROTOCOL_VERSION = "2"
"""TonConnect protocol version string."""

STANDARD_UNIVERSAL_LINK = "tc://"
"""Standard ``tc://`` universal link scheme."""


def add_path_to_url(url: str, path: str) -> str:
    """Append a path segment to a URL.

    :param url: Base URL.
    :param path: Path segment to append.
    :return: Combined URL.
    """
    return f"{_remove_url_last_slash(url)}/{path}"


def is_telegram_url(link: str | None = None) -> bool:
    """Check whether a link is a Telegram URL.

    :param link: URL to check, or ``None``.
    :return: ``True`` if the link uses ``tg://`` scheme or ``t.me`` host.
    """
    if not link:
        return False
    parsed = urlparse(link)
    return parsed.scheme == "tg" or parsed.hostname == "t.me"


def encode_telegram_url_parameters(parameters: str) -> str:
    """Encode query parameters for Telegram ``startapp`` format.

    :param parameters: Raw query string.
    :return: Telegram-safe encoded string.
    """
    return (
        parameters.replace(".", "%2E")
        .replace("-", "%2D")
        .replace("_", "%5F")
        .replace("&", "-")
        .replace("=", "__")
        .replace("%", "--")
    )


def decode_telegram_url_parameters(parameters: str) -> str:
    """Decode Telegram ``startapp`` parameters back to a query string.

    :param parameters: Telegram-encoded string.
    :return: Decoded query string.
    """
    return (
        parameters.replace("--", "%")
        .replace("__", "=")
        .replace("-", "&")
        .replace("%5F", "_")
        .replace("%2D", "-")
        .replace("%2E", ".")
    )


def generate_universal_link(
    universal_link: str,
    *,
    message: ConnectRequest,
    session_id: str,
    protocol_version: str | None = None,
    redirect_url: str | None = None,
) -> str:
    """Generate a TonConnect universal link.

    :param universal_link: Wallet universal link base URL.
    :param message: Connect request payload.
    :param session_id: Bridge session identifier.
    :param protocol_version: Protocol version string, or ``None`` for default.
    :param redirect_url: Post-connect redirect URL, or ``None``.
    :return: Ready-to-use universal link.
    """
    if protocol_version is None:
        protocol_version = TONCONNECT_PROTOCOL_VERSION
    if is_telegram_url(universal_link):
        return _generate_tg_universal_link(
            universal_link,
            message=message,
            session_id=session_id,
            protocol_version=protocol_version,
            redirect_url=redirect_url,
        )
    return _generate_regular_universal_link(
        universal_link,
        message=message,
        session_id=session_id,
        protocol_version=protocol_version,
        redirect_url=redirect_url,
    )


def _generate_regular_universal_link(
    universal_link: str,
    *,
    message: ConnectRequest,
    session_id: str,
    protocol_version: str,
    redirect_url: str | None = None,
) -> str:
    """Generate a regular (non-Telegram) universal link."""
    split = urlsplit(universal_link)
    q = dict(parse_qsl(split.query, keep_blank_values=True))

    q["v"] = protocol_version
    q["id"] = session_id
    q["r"] = message.dump_json()
    if redirect_url is not None:
        q["ret"] = redirect_url

    query = urlencode(q)

    if universal_link == STANDARD_UNIVERSAL_LINK:
        base = f"{split.scheme}://{split.path}"
        return f"{base}?{query}" if query else base

    comps = (split.scheme, split.netloc, split.path, query, split.fragment)
    return urlunsplit(comps)


def _generate_tg_universal_link(
    universal_link: str,
    *,
    message: ConnectRequest,
    session_id: str,
    protocol_version: str,
    redirect_url: str | None = None,
) -> str:
    """Generate a Telegram-specific universal link with ``startapp``."""
    url_to_wrap = _generate_regular_universal_link(
        "about:blank",
        message=message,
        session_id=session_id,
        protocol_version=protocol_version,
        redirect_url=redirect_url,
    )
    link_params = urlsplit(url_to_wrap).query

    startapp = "tonconnect-" + encode_telegram_url_parameters(link_params)

    updated = _convert_to_direct_link(universal_link)

    split = urlsplit(updated)
    q = dict(parse_qsl(split.query, keep_blank_values=True))
    q["startapp"] = startapp

    comps = (split.scheme, split.netloc, split.path, urlencode(q), split.fragment)
    return urlunsplit(comps)


def _convert_to_direct_link(universal_link: str) -> str:
    """Convert a Telegram attach-menu link to a direct bot link."""
    split = urlsplit(universal_link)
    q = dict(parse_qsl(split.query, keep_blank_values=True))

    if "attach" in q:
        q.pop("attach", None)
        path = split.path.rstrip("/") + "/start"
    else:
        path = split.path

    comps = (split.scheme, split.netloc, path, urlencode(q), split.fragment)
    return urlunsplit(comps)


def _remove_url_last_slash(url: str) -> str:
    """Strip trailing slash from a URL."""
    if url.endswith("/"):
        return url[:-1]
    return url



from .link import (
    STANDARD_UNIVERSAL_LINK,
    TONCONNECT_PROTOCOL_VERSION,
    add_path_to_url,
    decode_telegram_url_parameters,
    encode_telegram_url_parameters,
    generate_universal_link,
    is_telegram_url,
)
from .signing import (
    VerifySignData,
    VerifyTonProof,
    create_ton_proof_payload,
    verify_ton_proof_payload,
)
from .validation import (
    verify_send_transaction_support,
    verify_sign_data_support,
    verify_wallet_features,
    verify_wallet_network,
)
from .wallets_loader import AppWalletsLoader

__all__ = [
    "STANDARD_UNIVERSAL_LINK",
    "TONCONNECT_PROTOCOL_VERSION",
    "AppWalletsLoader",
    "VerifySignData",
    "VerifyTonProof",
    "add_path_to_url",
    "create_ton_proof_payload",
    "decode_telegram_url_parameters",
    "encode_telegram_url_parameters",
    "generate_universal_link",
    "is_telegram_url",
    "verify_send_transaction_support",
    "verify_sign_data_support",
    "verify_ton_proof_payload",
    "verify_wallet_features",
    "verify_wallet_network",
]



from .account import Account
from .app import (
    AppWallet,
    AppWallets,
    BridgeType,
    BridgeTypes,
    JSBridgeType,
    SSEBridgeType,
)
from .connection import (
    ActiveConnection,
    Connection,
    ConnectionSource,
    ConnectionSources,
    PendingConnection,
)
from .device import Device
from .dto import (
    SignDataPayloadDto,
    TonProofPayloadDto,
)
from .feature import (
    FeatureType,
    FeatureTypes,
    SendTransactionFeature,
    SignDataFeature,
)
from .message import IncomingMessage
from .payload import (
    BaseSignDataPayload,
    SendTransactionMessage,
    SendTransactionPayload,
    SignDataPayload,
    SignDataPayloadBinary,
    SignDataPayloadCell,
    SignDataPayloadText,
)
from .proof import (
    TonProofData,
    TonProofDomain,
)
from .request import (
    AppMessage,
    ConnectItem,
    ConnectRequest,
    DisconnectRpcRequest,
    RpcRequestBase,
    RpcRequests,
    RpcRequestWithParams,
    SendTransactionRpcRequest,
    SignDataRpcRequest,
    TonAddressItem,
    TonProofItem,
)
from .response import (
    ConnectEvent,
    ConnectEventError,
    ConnectEventItemReply,
    ConnectEventPayload,
    ConnectEventSuccess,
    DisconnectEvent,
    DisconnectEventError,
    DisconnectEventPayload,
    DisconnectEventSuccess,
    DisconnectRpcResponseSuccess,
    EventBase,
    EventErrorPayload,
    RpcBase,
    RpcRequestErrorPayload,
    RpcResponseSuccessBase,
    SendTransactionResult,
    SendTransactionRpcResponseSuccess,
    SignDataResult,
    SignDataRpcResponseSuccess,
    TonAddressItemReply,
    TonProofItemReply,
    WalletEvent,
    WalletMessage,
    WalletResponse,
    WalletResponseError,
    WalletResponseSuccess,
)
from .session import (
    BridgeProviderSession,
    SessionKeyPair,
)
from .wallet import Wallet

__all__ = [
    "Account",
    "ActiveConnection",
    "AppMessage",
    "AppWallet",
    "AppWallets",
    "BaseSignDataPayload",
    "BridgeProviderSession",
    "BridgeType",
    "BridgeTypes",
    "ConnectEvent",
    "ConnectEventError",
    "ConnectEventItemReply",
    "ConnectEventPayload",
    "ConnectEventSuccess",
    "ConnectItem",
    "ConnectRequest",
    "Connection",
    "ConnectionSource",
    "ConnectionSources",
    "Device",
    "DisconnectEvent",
    "DisconnectEventError",
    "DisconnectEventPayload",
    "DisconnectEventSuccess",
    "DisconnectRpcRequest",
    "DisconnectRpcResponseSuccess",
    "EventBase",
    "EventErrorPayload",
    "FeatureType",
    "FeatureTypes",
    "IncomingMessage",
    "JSBridgeType",
    "PendingConnection",
    "RpcBase",
    "RpcRequestBase",
    "RpcRequestErrorPayload",
    "RpcRequestWithParams",
    "RpcRequests",
    "RpcResponseSuccessBase",
    "SSEBridgeType",
    "SendTransactionFeature",
    "SendTransactionMessage",
    "SendTransactionPayload",
    "SendTransactionResult",
    "SendTransactionRpcRequest",
    "SendTransactionRpcResponseSuccess",
    "SessionKeyPair",
    "SignDataFeature",
    "SignDataPayload",
    "SignDataPayloadBinary",
    "SignDataPayloadCell",
    "SignDataPayloadDto",
    "SignDataPayloadText",
    "SignDataResult",
    "SignDataRpcRequest",
    "SignDataRpcResponseSuccess",
    "TonAddressItem",
    "TonAddressItemReply",
    "TonProofData",
    "TonProofDomain",
    "TonProofItem",
    "TonProofItemReply",
    "TonProofPayloadDto",
    "Wallet",
    "WalletEvent",
    "WalletMessage",
    "WalletResponse",
    "WalletResponseError",
    "WalletResponseSuccess",
]



import asyncio
import time
import typing as t

from pydantic import TypeAdapter, ValidationError

from ..exceptions import TonConnectError
from ..models.connection import (
    ActiveConnection,
    Connection,
    PendingConnection,
)
from ..storage import StorageProtocol


class ProviderStorage:
    """High-level storage wrapper for TonConnect provider state."""

    _key_prefix = "ton-connect-storage"
    _connection_adapter: t.ClassVar[TypeAdapter[Connection]] = TypeAdapter(Connection)

    def __init__(self, storage: StorageProtocol, session_key: str) -> None:
        """Initialize provider storage.

        :param storage: Underlying key-value storage.
        :param session_key: Unique session identifier for key namespacing.
        """
        self._storage = storage
        self._lock = asyncio.Lock()
        self._key = f"{self._key_prefix}::{session_key}"

    async def _get_connection(self) -> Connection | None:
        """Load and validate connection from storage, or ``None``."""
        raw = await self._storage.get_item(self._key)
        if not isinstance(raw, dict):
            return None

        try:
            conn = self._connection_adapter.validate_python(raw)
        except ValidationError:
            await self.remove_connection()
            return None

        if isinstance(conn, PendingConnection) and conn.is_expired():
            await self.remove_connection()
            return None

        return conn

    async def store_connection(self, connection: Connection) -> None:
        """Persist a connection (pending or active).

        :param connection: Connection to store.
        """
        if isinstance(connection, PendingConnection) and connection.created_at is None:
            connection = connection.model_copy(update={"created_at": int(time.time())})

        conn = connection.dump()
        await self._storage.set_item(self._key, conn)

    async def remove_connection(self) -> None:
        """Remove the stored connection."""
        await self._storage.remove_item(self._key)

    async def get_connection(self) -> Connection:
        """Load the stored connection.

        :return: Active or pending connection.
        :raises TonConnectError: If no connection is stored.
        """
        conn = await self._get_connection()
        if not conn:
            raise TonConnectError("Trying to read connection source while nothing is stored")

        return conn

    async def get_pending_connection(self) -> PendingConnection:
        """Load the stored pending connection.

        :return: Pending connection.
        :raises TonConnectError: If no pending connection is stored.
        """
        conn = await self._get_connection()
        if not conn:
            raise TonConnectError("Trying to read connection source while nothing is stored")

        if not isinstance(conn, PendingConnection):
            raise TonConnectError("Trying to read inflight connection while connection is stored")

        return conn

    async def store_last_wallet_event_id(self, event_id: int | str) -> None:
        """Persist the last wallet event ID.

        :param event_id: Monotonic event identifier.
        """
        async with self._lock:
            conn = await self._get_connection()
            if isinstance(conn, ActiveConnection):
                conn = conn.model_copy(update={"last_wallet_event_id": event_id})
                await self.store_connection(conn)

    async def get_last_wallet_event_id(self) -> int | str | None:
        """Return the last wallet event ID, or ``None``."""
        conn = await self._get_connection()
        if isinstance(conn, ActiveConnection):
            return conn.last_wallet_event_id
        return None

    async def increase_next_rpc_request_id(self) -> None:
        """Increment the next RPC request ID in storage."""
        async with self._lock:
            conn = await self._get_connection()
            if isinstance(conn, ActiveConnection):
                last_id = int(conn.next_rpc_request_id or 0)
                conn = conn.model_copy(update={"next_rpc_request_id": last_id + 1})
                await self.store_connection(conn)

    async def get_next_rpc_request_id(self) -> int:
        """Return the next RPC request ID (0 if not active)."""
        conn = await self._get_connection()
        if isinstance(conn, ActiveConnection):
            return int(conn.next_rpc_request_id or 0)
        return 0

    async def store_last_event_id(self, last_event_id: str) -> None:
        """Persist the last SSE event ID.

        :param last_event_id: SSE event identifier string.
        """
        async with self._lock:
            conn = await self._get_connection()
            if isinstance(conn, ActiveConnection):
                conn = conn.model_copy(update={"last_event_id": last_event_id})
                await self.store_connection(conn)

    async def get_last_event_id(self) -> str | None:
        """Return the last SSE event ID, or ``None``."""
        conn = await self._get_connection()
        if isinstance(conn, ActiveConnection):
            return conn.last_event_id
        return None



from ._types import (
    A,
    BaseModel,
    ChainId,
    TonAddress,
    TonPublicKey,
    WalletStateInit,
)


class Account(BaseModel):
    """Connected wallet account data."""

    address: TonAddress
    """Wallet address."""
    network: ChainId
    """Network identifier."""
    public_key: TonPublicKey = A("publicKey")
    """Wallet public key."""
    state_init: WalletStateInit = A("walletStateInit")
    """Wallet ``StateInit``."""



import abc
import hashlib
import hmac
import secrets
import time
import typing as t
import zlib

from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey
from ton_core import (
    CONTRACT_CODES,
    Address,
    BaseWalletData,
    Cell,
    ContractVersion,
    NetworkGlobalID,
    PublicKey,
    SignatureDomain,
    StateInit,
    WalletV1Data,
    WalletV2Data,
    WalletV3Data,
    WalletV4Data,
    WalletV5BetaData,
    WalletV5Data,
    begin_cell,
    encode_dns_name,
    to_cell,
)

from ..models import (
    SignDataPayloadBinary,
    SignDataPayloadCell,
    SignDataPayloadDto,
    SignDataPayloadText,
    TonProofPayloadDto,
)

__all__ = [
    "VerifySignData",
    "VerifyTonProof",
    "create_ton_proof_payload",
    "verify_ton_proof_payload",
]


def _code(version: ContractVersion) -> Cell:
    return to_cell(CONTRACT_CODES[version])


_WALLET_DATA_MODELS: dict[Cell, type[BaseWalletData]] = {
    _code(ContractVersion.WalletV1R1): WalletV1Data,
    _code(ContractVersion.WalletV1R2): WalletV1Data,
    _code(ContractVersion.WalletV1R3): WalletV1Data,
    _code(ContractVersion.WalletV2R1): WalletV2Data,
    _code(ContractVersion.WalletV2R2): WalletV2Data,
    _code(ContractVersion.WalletV3R1): WalletV3Data,
    _code(ContractVersion.WalletV3R2): WalletV3Data,
    _code(ContractVersion.WalletV4R1): WalletV4Data,
    _code(ContractVersion.WalletV4R2): WalletV4Data,
    _code(ContractVersion.WalletV5R1): WalletV5Data,
    _code(ContractVersion.WalletV5Beta): WalletV5BetaData,
}

_TDTO = t.TypeVar("_TDTO", TonProofPayloadDto, SignDataPayloadDto)
_GetWalletPublicKey = t.Callable[[Address], t.Awaitable[PublicKey | None]]


def _u32_be(value: int) -> bytes:
    """Encode unsigned 32-bit integer (big-endian)."""
    return int(value).to_bytes(4, "big")


def _u64_be(value: int) -> bytes:
    """Encode unsigned 64-bit integer (big-endian)."""
    return int(value).to_bytes(8, "big")


def _u32_le(value: int) -> bytes:
    """Encode unsigned 32-bit integer (little-endian)."""
    return int(value).to_bytes(4, "little")


def _u64_le(value: int) -> bytes:
    """Encode unsigned 64-bit integer (little-endian)."""
    return int(value).to_bytes(8, "little")


def _i32_be(value: int) -> bytes:
    """Encode signed 32-bit integer (big-endian)."""
    return int(value).to_bytes(4, "big", signed=True)


class _BaseVerify(abc.ABC, t.Generic[_TDTO]):
    """Base verifier for TON Connect signed payload DTOs."""

    TON_CONNECT_PREFIX: bytes

    def __init__(self, dto: _TDTO) -> None:
        """Initialize verifier with the given DTO."""
        self.dto: _TDTO = dto

    @property
    @abc.abstractmethod
    def address(self) -> Address:
        """Wallet address that must be proven."""

    @property
    @abc.abstractmethod
    def domain(self) -> str:
        """dApp domain that must match the signed domain."""

    @property
    @abc.abstractmethod
    def public_key(self) -> PublicKey:
        """Public key declared by the wallet."""

    @property
    @abc.abstractmethod
    def signature(self) -> bytes:
        """Ed25519 signature bytes."""

    @property
    @abc.abstractmethod
    def signing_message(self) -> bytes:
        """Canonical message bytes the wallet signs."""

    @property
    @abc.abstractmethod
    def timestamp(self) -> int:
        """Unix timestamp when the wallet produced the signature."""

    @property
    @abc.abstractmethod
    def wallet_state_init(self) -> StateInit:
        """Wallet ``StateInit`` used to derive address and public key."""

    async def verify(
        self,
        allowed_domains: list[str],
        valid_auth_time: int = 15 * 60,
        get_wallet_public_key: _GetWalletPublicKey | None = None,
    ) -> bool:
        """Verify payload DTO authenticity.

        :param allowed_domains: Permitted domain strings.
        :param valid_auth_time: Max age of signature in seconds.
        :param get_wallet_public_key: Async resolver for unknown wallet codes, or ``None``.
        :return: ``True`` if all checks pass.
        :raises BadSignatureError: If any validation step fails.
        """
        await self._resolve_public_key(
            wallet_state_init=self.wallet_state_init,
            address=self.address,
            network=self.dto.network,
            wanted_public_key=self.public_key,
            get_wallet_public_key=get_wallet_public_key,
        )

        self._ensure_address_matches(self.address, self.wallet_state_init)
        self._ensure_timestamp_valid(self.timestamp, valid_auth_time)
        self._ensure_domain_allowed(self.domain, allowed_domains)

        verify_key = VerifyKey(self.public_key.as_bytes)
        data = self.signing_message
        domain = SignatureDomain(self.dto.network)

        if domain.is_l2:
            try:
                return verify_key.verify(domain.data_to_sign(data), self.signature) is not None
            except BadSignatureError:
                pass

        return verify_key.verify(data, self.signature) is not None

    @classmethod
    def _ensure_domain_allowed(cls, domain: str, allowed_domains: list[str]) -> None:
        """Ensure the domain is in the allow-list.

        :param domain: Domain from the signed payload.
        :param allowed_domains: Backend allow-list.
        :raises BadSignatureError: If the domain is not allowed.
        """
        if domain not in allowed_domains:
            raise BadSignatureError("Domain not allowed")

    @classmethod
    def _ensure_timestamp_valid(cls, timestamp: int, valid_auth_time: int) -> None:
        """Ensure the timestamp is within the allowed window.

        :param timestamp: Signed unix timestamp.
        :param valid_auth_time: Max age in seconds.
        :raises BadSignatureError: If timestamp is out of range.
        """
        now = int(time.time())
        if timestamp < now - valid_auth_time:
            raise BadSignatureError("Signature expired")
        if timestamp > now + 60:
            raise BadSignatureError("Invalid timestamp")

    @classmethod
    def _ensure_address_matches(cls, address: Address, state_init: StateInit) -> None:
        """Ensure address matches ``StateInit``.

        :param address: Wallet address from payload.
        :param state_init: Wallet ``StateInit`` from payload.
        :raises BadSignatureError: If derivation fails or address mismatches.
        """
        try:
            st_cell = state_init.serialize()
            derived = Address((address.wc, st_cell.hash))
        except Exception as e:
            raise BadSignatureError("Cannot validate address against state_init") from e

        if derived != address:
            raise BadSignatureError("Address does not match wallet_state_init")

    @classmethod
    async def _resolve_public_key(
        cls,
        wallet_state_init: StateInit,
        address: Address,
        network: NetworkGlobalID,
        wanted_public_key: PublicKey,
        get_wallet_public_key: _GetWalletPublicKey | None,
    ) -> PublicKey:
        """Resolve and validate the wallet public key.

        :param wallet_state_init: Wallet ``StateInit``.
        :param address: Wallet address (for external resolver).
        :param network: Network identifier for V5 wallet deserialization.
        :param wanted_public_key: Public key declared by the wallet.
        :param get_wallet_public_key: Async resolver, or ``None``.
        :return: Resolved public key.
        :raises BadSignatureError: If key cannot be resolved or mismatches.
        """
        public_key = cls._try_parse_public_key(wallet_state_init, network)
        if public_key is None and get_wallet_public_key is not None:
            public_key = await get_wallet_public_key(address)

        if public_key is None:
            raise BadSignatureError("Public key not found")
        if public_key != wanted_public_key:
            raise BadSignatureError("Public key mismatch")
        return public_key

    @classmethod
    def _try_parse_public_key(
        cls,
        state_init: StateInit,
        network: NetworkGlobalID,
    ) -> PublicKey | None:
        """Try to extract a public key from ``StateInit``.

        :param state_init: Wallet ``StateInit``.
        :param network: Network identifier for V5 wallet deserialization.
        :return: Extracted public key, or ``None``.
        """
        code, data = state_init.code, state_init.data
        if code is None or data is None:
            return None

        data_model = _WALLET_DATA_MODELS.get(code)
        if data_model is not None:
            cs = data.begin_parse()
            if data_model is WalletV5Data:
                v5_network = NetworkGlobalID.MAINNET if network == NetworkGlobalID.TETRA else network
                wallet_data = data_model.deserialize(cs, v5_network)
            else:
                wallet_data = data_model.deserialize(cs)
            return t.cast("PublicKey", wallet_data.public_key)

        return None


class VerifySignData(_BaseVerify[SignDataPayloadDto]):
    """Verifier for TON Connect ``signData`` payload DTOs."""

    TON_CONNECT_PREFIX = b"ton-connect/sign-data/"

    @property
    def address(self) -> Address:
        """Wallet address."""
        return self.dto.address

    @property
    def domain(self) -> str:
        """dApp domain."""
        return self.dto.domain

    @property
    def public_key(self) -> PublicKey:
        """Wallet public key."""
        return self.dto.public_key

    @property
    def signature(self) -> bytes:
        """Ed25519 signature bytes."""
        return self.dto.signature.as_bytes

    @property
    def signing_message(self) -> bytes:
        """Build the canonical signing message for ``signData``.

        :raises BadSignatureError: If payload type is unsupported.
        """
        if isinstance(self.dto.payload, SignDataPayloadCell):
            schema = self.dto.payload.tlb_schema.encode()
            schema_hash = zlib.crc32(schema) & 0xFFFFFFFF
            domain = encode_dns_name(self.domain)
            domain_cell = begin_cell().store_snake_bytes(domain).end_cell()
            return self._build_cell_message(schema_hash, domain_cell)

        if isinstance(self.dto.payload, SignDataPayloadText):
            prefix = b"txt"
            content = self.dto.payload.text.encode()
            return self._build_text_binary_message(prefix, content)

        if isinstance(self.dto.payload, SignDataPayloadBinary):
            prefix = b"bin"
            content = self.dto.payload.raw_bytes
            return self._build_text_binary_message(prefix, content)

        raise BadSignatureError("Unsupported payload type")

    @property
    def timestamp(self) -> int:
        """Signature unix timestamp."""
        return self.dto.timestamp

    @property
    def wallet_state_init(self) -> StateInit:
        """Wallet ``StateInit``."""
        return self.dto.wallet_state_init

    def _build_cell_message(self, schema_hash: int, domain_cell: Cell) -> bytes:
        """Build signing message for cell payload.

        :param schema_hash: CRC32 of TL-B schema string.
        :param domain_cell: ``Cell`` with snake-encoded DNS name.
        :return: 32-byte cell hash.
        """
        assert isinstance(self.dto.payload, SignDataPayloadCell)
        cell = begin_cell()
        cell.store_uint(0x75569022, 32)
        cell.store_uint(schema_hash, 32)
        cell.store_uint(self.timestamp, 64)
        cell.store_address(self.address)
        cell.store_ref(domain_cell)
        cell.store_ref(self.dto.payload.cell)
        return cell.end_cell().hash

    def _build_text_binary_message(self, prefix: bytes, content: bytes) -> bytes:
        """Build signing message for text or binary payload.

        :param prefix: Type tag (``b"txt"`` or ``b"bin"``).
        :param content: Content bytes.
        :return: 32-byte SHA-256 digest.
        """
        msg = bytearray()
        msg.extend(b"\xff\xff")
        msg.extend(self.TON_CONNECT_PREFIX)
        msg.extend(_i32_be(self.address.wc))
        msg.extend(self.address.hash_part)

        domain_bytes = self.domain.encode()
        msg.extend(_u32_be(len(domain_bytes)))
        msg.extend(domain_bytes)

        msg.extend(_u64_be(self.timestamp))
        msg.extend(prefix)
        msg.extend(_u32_be(len(content)))
        msg.extend(content)
        return hashlib.sha256(bytes(msg)).digest()


class VerifyTonProof(_BaseVerify[TonProofPayloadDto]):
    """Verifier for TON Connect ``ton_proof`` payload DTOs."""

    TON_CONNECT_PREFIX = b"ton-connect"

    @property
    def address(self) -> Address:
        """Wallet address."""
        return self.dto.address

    @property
    def domain(self) -> str:
        """dApp domain."""
        return self.dto.proof.domain.value

    @property
    def public_key(self) -> PublicKey:
        """Wallet public key."""
        return self.dto.public_key

    @property
    def signature(self) -> bytes:
        """Ed25519 signature bytes."""
        return self.dto.proof.signature.as_bytes

    @property
    def signing_message(self) -> bytes:
        """Compute canonical signing hash for TON Proof."""
        msg = self._build_proof_message()
        full = bytearray()
        full.extend(b"\xff\xff")
        full.extend(self.TON_CONNECT_PREFIX)
        full.extend(hashlib.sha256(msg).digest())
        return hashlib.sha256(bytes(full)).digest()

    @property
    def timestamp(self) -> int:
        """Proof unix timestamp."""
        return self.dto.proof.timestamp

    @property
    def wallet_state_init(self) -> StateInit:
        """Wallet ``StateInit``."""
        return self.dto.wallet_state_init

    def _build_proof_message(self) -> bytes:
        """Build raw ``ton-proof-item-v2/`` message bytes.

        :return: Raw message bytes (not hashed).
        """
        msg = bytearray()
        msg.extend(b"ton-proof-item-v2/")
        msg.extend(_i32_be(self.address.wc))
        msg.extend(self.address.hash_part)
        msg.extend(_u32_le(self.dto.proof.domain.length_bytes))
        msg.extend(self.domain.encode())
        msg.extend(_u64_le(self.timestamp))
        msg.extend(self.dto.proof.payload.encode())
        return bytes(msg)


def create_ton_proof_payload(secret_key: str, ttl: int = 15 * 60) -> str:
    """Create backend-generated challenge payload for TON Proof.

    :param secret_key: Backend secret key for HMAC.
    :param ttl: Validity period in seconds.
    :return: Hex-encoded payload.
    """
    exp = int(time.time()) + int(ttl)
    payload = secrets.token_bytes(32) + exp.to_bytes(8, "big")
    sig = hmac.new(secret_key.encode(), payload, hashlib.sha256).digest()
    return (payload + sig).hex()


def verify_ton_proof_payload(secret_key: str, ton_proof_payload: str) -> bool:
    """Verify backend-generated TON Proof challenge payload.

    :param secret_key: Backend secret key used in ``create_ton_proof_payload``.
    :param ton_proof_payload: Hex payload from wallet proof response.
    :return: ``True`` if payload is valid.
    :raises BadSignatureError: If payload is malformed, signature invalid, or expired.
    """
    try:
        raw = bytes.fromhex(ton_proof_payload)
    except Exception as e:
        raise BadSignatureError("Invalid payload encoding") from e

    if len(raw) != 72:
        raise BadSignatureError("Invalid payload length")

    payload, sig = raw[:40], raw[40:]
    expected = hmac.new(secret_key.encode(), payload, hashlib.sha256).digest()
    if not hmac.compare_digest(sig, expected):
        raise BadSignatureError("Payload signature mismatch")

    exp = int.from_bytes(payload[32:40], "big")
    if int(time.time()) > exp + 60:
        raise BadSignatureError("Payload expired")
    return True



from codecs import getincrementaldecoder

from .message import EventMessage

_UTF8_DECODER = getincrementaldecoder("utf-8")


class EventDecoder:
    """Incremental SSE stream decoder."""

    def __init__(self) -> None:
        """Initialize decoder with empty buffer."""
        self._decoder = _UTF8_DECODER()
        self._buffer: str = ""

    def reset(self) -> None:
        """Reset decoder and buffer state."""
        self._decoder = _UTF8_DECODER()
        self._buffer = ""

    def feed(self, chunk: bytes) -> list[EventMessage]:
        """Decode a chunk and return any complete events.

        :param chunk: Raw bytes from the SSE stream.
        :return: List of parsed events (maybe empty).
        """
        text = self._decoder.decode(chunk)
        if not text:
            return []

        text = text.replace("\r\n", "\n").replace("\r", "\n")
        self._buffer += text

        buf = self._buffer
        events: list[EventMessage] = []

        while True:
            sep = buf.find("\n\n")
            if sep == -1:
                break

            raw_event = buf[:sep]
            buf = buf[sep + 2 :]

            if raw_event.strip():
                events.append(EventMessage.parse(raw_event))

        self._buffer = buf
        return events

    def flush(self) -> list[EventMessage]:
        """Flush remaining buffer and return any trailing event.

        :return: List with at most one event.
        """
        raw = self._buffer
        self.reset()

        if raw.strip():
            return [EventMessage.parse(raw)]
        return []



from enum import Enum


class ReadyState(int, Enum):
    """SSE connection state."""

    CONNECTING = 0
    """Connection in progress (0)."""
    OPEN = 1
    """Connected and receiving events (1)."""
    CLOSED = 2
    """Connection closed (2)."""



import base64
import time
import typing as t

from pydantic import Field, field_serializer, field_validator
from ton_core import Cell, TextCommentBody

from ._types import (
    A,
    BaseModel,
    BocCell,
    OptionalBocCell,
    OptionalChainId,
    OptionalTonAddress,
    OptionalWalletStateInit,
    TonAddress,
)


class SendTransactionMessage(BaseModel):
    """Single outgoing message within a ``sendTransaction`` request."""

    address: TonAddress
    """Destination address."""
    amount: int
    """Transfer amount in nanotons."""
    state_init: OptionalWalletStateInit = A("stateInit", default=None)
    """Contract ``StateInit``, or ``None``."""
    payload: OptionalBocCell | str | None = None
    """Message body ``Cell``, or ``None``."""
    extra_currency: dict[int, str] | None = A("extraCurrency", default=None)
    """Extra currency map, or ``None``."""

    @field_validator("amount", mode="before")
    @classmethod
    def _v_amount(cls, v: t.Any) -> int:
        if isinstance(v, int):
            return v
        return int(v)

    @field_serializer("amount")
    def _s_amount(self, v: int) -> str:
        return str(v)

    @field_validator("payload", mode="before")
    @classmethod
    def _v_payload(cls, v: t.Any) -> Cell | None:
        if isinstance(v, str):
            return TextCommentBody(v).serialize()
        return t.cast("Cell | None", v)


class SendTransactionPayload(BaseModel):
    """Payload for a ``sendTransaction`` RPC request."""

    network: OptionalChainId = None
    """Target network, or ``None``."""
    from_address: OptionalTonAddress = A("from", default=None)
    """Sender address override, or ``None``."""
    valid_until: int = A("validUntil", default=None)
    """Expiry unix timestamp."""
    messages: list[SendTransactionMessage] = Field(default_factory=list)
    """Outgoing messages."""

    @field_validator("valid_until", mode="before")
    @classmethod
    def _v_valid_until(cls, v: t.Any) -> int:
        if v is None:
            return int(time.time()) + 5 * 60
        return int(v)


class BaseSignDataPayload(BaseModel):
    """Common fields for all ``signData`` payload types."""

    network: OptionalChainId = None
    """Target network, or ``None``."""
    from_address: OptionalTonAddress = A("from", default=None)
    """Wallet address, or ``None``."""


class SignDataPayloadText(BaseSignDataPayload):
    """Text payload for ``signData``."""

    type: t.Literal["text"] = "text"
    """Payload type literal."""
    text: str
    """UTF-8 text to sign."""


class SignDataPayloadBinary(BaseSignDataPayload):
    """Binary payload for ``signData``."""

    type: t.Literal["binary"] = "binary"
    """Payload type literal."""
    raw_bytes: bytes = A("bytes")
    """Raw bytes to sign."""

    @field_validator("raw_bytes", mode="before")
    @classmethod
    def _v_raw_bytes(cls, v: t.Any) -> bytes:
        if isinstance(v, bytes):
            return v
        s = str(v).strip()
        padded = s + "=" * (-len(s) % 4)
        return base64.b64decode(padded, validate=True)

    @field_serializer("raw_bytes")
    def _s_raw_bytes(self, v: bytes) -> str:
        return base64.b64encode(v).decode("ascii")


class SignDataPayloadCell(BaseSignDataPayload):
    """Cell payload for ``signData``."""

    type: t.Literal["cell"] = "cell"
    """Payload type literal."""
    tlb_schema: str = A("schema")
    """TL-B schema string."""
    cell: BocCell
    """Signing payload ``Cell``."""


SignDataPayload: t.TypeAlias = t.Annotated[
    SignDataPayloadBinary | SignDataPayloadCell | SignDataPayloadText,
    Field(discriminator="type"),
]



import json
import time
import typing as t
from contextlib import suppress
from pathlib import Path

from pydantic import TypeAdapter
from ton_core import load_json

from ..exceptions import FetchWalletsError
from ..models.app import AppWallet, AppWallets

_WALLETS_ADAPTER = TypeAdapter(AppWallets)

DEFAULT_WALLETS_LIST_SOURCE = "https://config.ton.org/wallets-v2.json"
"""Default URL for the wallets catalogue."""

DEFAULT_WALLETS_FALLBACK_FILE = "wallets-v2.json"
"""Default filename for the local wallets fallback."""


class AppWalletsLoader:
    """Wallets catalogue loader with memory cache, remote fetch, and file fallback."""

    def __init__(
        self,
        source: str | Path = DEFAULT_WALLETS_LIST_SOURCE,
        include_wallets: list[str] | None = None,
        exclude_wallets: list[str] | None = None,
        order_wallets: list[str] | None = None,
        require_sign_data: bool = False,
        fallback_path: Path | None = None,
        cache_ttl: float | None = None,
        timeout: float = 5.0,
    ) -> None:
        """Initialize the wallet catalogue loader.

        :param source: Remote URL or local path to wallets JSON.
        :param include_wallets: App names to include, or ``None`` for all.
        :param exclude_wallets: App names to exclude, or ``None``.
        :param order_wallets: Desired ordering of app names, or ``None``.
        :param require_sign_data: If ``True``, only include wallets that support
            the ``SignData`` feature.
        :param fallback_path: Local fallback file path, or ``None``.
        :param cache_ttl: Cache lifetime in seconds, or ``None`` for no expiry.
        :param timeout: HTTP fetch timeout in seconds.
        """
        self._source = source
        self._path = self._resolve_fallback_path(fallback_path)

        self._include = set(include_wallets) if include_wallets else None
        self._exclude = set(exclude_wallets) if exclude_wallets else None
        self._order = order_wallets
        self._require_sign_data = require_sign_data

        self._cache_ttl = cache_ttl
        self._timeout = timeout

        self._cache: list[AppWallet] | None = None
        self._cache_data: dict[str, AppWallet] = {}
        self._cached_at: float = 0.0

    def get_wallets(self) -> list[AppWallet]:
        """Return the wallets list, fetching and caching as needed.

        :return: Filtered and ordered wallet list.
        :raises FetchWalletsError: If wallets cannot be loaded.
        """
        if self._cache is not None and not self._expired():
            return self._cache

        raw = self._fetch_remote()
        if raw is None:
            raw = self._load_fallback()
        if raw is None:
            raise FetchWalletsError("Wallets catalog unavailable")

        wallets = self._process(raw)

        self._cache = wallets
        self._cache_data = {w.app_name: w for w in wallets}
        self._cached_at = time.time()

        return wallets

    def get_wallet(self, app_name: str) -> AppWallet | None:
        """Return a single wallet by app name, or ``None``.

        :param app_name: Wallet application name.
        :return: Wallet descriptor, or ``None``.
        """
        if self._cache is None or self._expired():
            self.get_wallets()
        return self._cache_data.get(app_name)

    def _process(self, raw: t.Any) -> list[AppWallet]:
        """Parse, filter, and sort raw wallet data."""
        wallets = self._parse(raw)
        wallets = self._filter(wallets)
        wallets = self._sort(wallets)
        return wallets  # noqa: RET504

    @staticmethod
    def _parse(raw: t.Any) -> list[AppWallet]:
        """Parse and validate raw JSON data.

        :param raw: Raw JSON-decoded data.
        :return: Wallets with a bridge URL.
        :raises FetchWalletsError: If data is invalid.
        """
        try:
            data = _WALLETS_ADAPTER.validate_python(raw)
        except Exception as e:
            raise FetchWalletsError(f"Invalid wallets data: {e}") from e

        return [w for w in data if w.bridge_url]

    def _filter(self, wallets: list[AppWallet]) -> list[AppWallet]:
        """Apply include/exclude/feature filters."""
        if self._include is not None:
            wallets = [w for w in wallets if w.app_name in self._include]
        if self._exclude is not None:
            wallets = [w for w in wallets if w.app_name not in self._exclude]
        if self._require_sign_data:
            wallets = [w for w in wallets if any(f.name == "SignData" for f in w.features)]

        return wallets

    def _sort(self, wallets: list[AppWallet]) -> list[AppWallet]:
        """Sort wallets by the configured order."""
        if not self._order:
            return wallets

        order_map: dict[str, int] = {name: i for i, name in enumerate(self._order)}
        default = len(order_map)

        wallets.sort(key=lambda w: order_map.get(w.app_name, default))
        return wallets

    def _fetch_remote(self) -> t.Any | None:
        """Fetch wallets JSON from the remote source, or ``None`` on failure."""
        try:
            raw = load_json(str(self._source), timeout=self._timeout)
            self._save_fallback(raw)
            return raw
        except Exception:
            return None

    def _load_fallback(self) -> t.Any | None:
        """Load wallets from the local fallback file, or ``None``."""
        if self._path is None:
            return None
        try:
            return json.loads(self._path.read_text())
        except Exception:
            return None

    def _save_fallback(self, raw: t.Any) -> None:
        """Save fetched data to the fallback file."""
        if self._path is None:
            return
        with suppress(Exception):
            self._path.parent.mkdir(parents=True, exist_ok=True)
            self._path.write_text(json.dumps(raw))

    def _expired(self) -> bool:
        """Check whether the cache has expired."""
        if self._cache_ttl is None:
            return False
        return (time.time() - self._cached_at) > self._cache_ttl

    @staticmethod
    def _resolve_fallback_path(path: Path | None) -> Path | None:
        """Resolve fallback path, appending default filename if directory.

        :param path: User-provided path, or ``None``.
        :return: Resolved file path, or ``None``.
        """
        if path is None:
            return None
        path = Path(path)
        if path.is_dir():
            return path / DEFAULT_WALLETS_FALLBACK_FILE
        return path



import typing as t

from pydantic import Field, RootModel
from ton_core import Cell, MessageAny, normalize_hash

from ._types import (
    A,
    BaseModel,
    Binary64,
    ChainId,
    TonAddress,
    TonPublicKey,
    WalletStateInit,
)
from .device import Device
from .payload import SignDataPayload
from .proof import TonProofData


class EventBase(BaseModel):
    """Base for wallet events with a monotonic ID."""

    id: int | str | None = None
    """Monotonic event identifier."""


class TonAddressItemReply(BaseModel):
    """Wallet reply containing TON address and keys."""

    name: t.Literal["ton_addr"]
    """Item type literal."""
    address: TonAddress
    """Wallet address."""
    network: ChainId
    """Network identifier."""
    state_init: WalletStateInit = A("walletStateInit")
    """Wallet ``StateInit``."""
    public_key: TonPublicKey = A("publicKey")
    """Wallet public key."""


class TonProofItemReply(BaseModel):
    """Wallet reply containing TON Proof data."""

    name: t.Literal["ton_proof"]
    """Item type literal."""
    proof: TonProofData | None = None
    """Proof data, or ``None``."""


ConnectEventItemReply: t.TypeAlias = t.Annotated[
    TonAddressItemReply | TonProofItemReply,
    Field(discriminator="name"),
]


class ConnectEventPayload(BaseModel):
    """Payload of a successful connect event."""

    items: list[ConnectEventItemReply]
    """Connect reply items."""
    device: Device
    """Wallet device information."""


class DisconnectEventPayload(BaseModel):
    """Payload of a successful disconnect event."""

    data: dict[str, t.Any] = Field(default_factory=dict)
    """Additional data (typically empty)."""


class EventErrorPayload(BaseModel):
    """Payload of an event error."""

    code: int
    """Error code."""
    message: str
    """Error message."""


class ConnectEventSuccess(EventBase):
    """Successful wallet connect event."""

    event: t.Literal["connect"]
    """Event type literal."""
    payload: ConnectEventPayload
    """Connect event payload."""


class ConnectEventError(EventBase):
    """Failed wallet connect event."""

    event: t.Literal["connect_error"]
    """Event type literal."""
    payload: EventErrorPayload
    """Error payload."""


class DisconnectEventSuccess(EventBase):
    """Successful wallet disconnect event."""

    event: t.Literal["disconnect"]
    """Event type literal."""
    payload: dict[str, t.Any] = Field(default_factory=dict)
    """Additional data (typically empty)."""


class DisconnectEventError(EventBase):
    """Failed wallet disconnect event."""

    event: t.Literal["disconnect_error"]
    """Event type literal."""
    payload: EventErrorPayload
    """Error payload."""


ConnectEvent: t.TypeAlias = ConnectEventSuccess | ConnectEventError

DisconnectEvent: t.TypeAlias = DisconnectEventSuccess | DisconnectEventError

WalletEvent: t.TypeAlias = ConnectEvent | DisconnectEvent


class RpcBase(BaseModel):
    """Base for wallet RPC responses."""

    id: str
    """Request identifier string."""


class RpcRequestErrorPayload(BaseModel):
    """Error payload in an RPC response."""

    code: int
    """Error code."""
    message: str
    """Error message."""
    data: t.Any | None = None
    """Additional error data, or ``None``."""


class WalletResponseError(RpcBase):
    """Wallet RPC error response."""

    error: RpcRequestErrorPayload
    """Error payload."""


class SendTransactionResult(RootModel[str]):
    """Result of a ``sendTransaction`` RPC call."""

    @property
    def boc(self) -> str:
        """Base64-encoded BOC string."""
        return self.root

    def to_cell(self) -> Cell:
        """Deserialize the BOC to a ``Cell``.

        :return: Parsed ``Cell``.
        """
        return Cell.one_from_boc(self.root)

    @property
    def normalized_hash(self) -> str:
        """Normalized message hash for tracking."""
        msg = MessageAny.deserialize(self.to_cell().begin_parse())
        return normalize_hash(msg)


class RpcResponseSuccessBase(RpcBase):
    """Base for successful RPC responses."""


class SignDataResult(BaseModel):
    """Result of a ``signData`` RPC call."""

    signature: Binary64
    """Ed25519 signature (64 bytes)."""
    address: TonAddress
    """Signer wallet address."""
    timestamp: int
    """Signing unix timestamp."""
    domain: str
    """dApp domain."""
    payload: SignDataPayload
    """Signed data payload."""


class SendTransactionRpcResponseSuccess(RpcResponseSuccessBase):
    """Successful ``sendTransaction`` RPC response."""

    result: SendTransactionResult
    """Transaction result."""


class DisconnectRpcResponseSuccess(RpcResponseSuccessBase):
    """Successful ``disconnect`` RPC response."""

    result: dict[str, t.Any] = Field(default_factory=dict)
    """Additional data (typically empty)."""


class SignDataRpcResponseSuccess(RpcResponseSuccessBase):
    """Successful ``signData`` RPC response."""

    result: SignDataResult
    """Sign data result."""


WalletResponseSuccess: t.TypeAlias = (
    DisconnectRpcResponseSuccess | SendTransactionRpcResponseSuccess | SignDataRpcResponseSuccess
)

WalletResponse: t.TypeAlias = WalletResponseSuccess | WalletResponseError

WalletMessage: t.TypeAlias = WalletEvent | WalletResponse



import typing as t

from ton_core import NetworkGlobalID

from ..exceptions import (
    WalletMissingRequiredFeaturesError,
    WalletNotSupportFeatureError,
    WalletWrongNetworkError,
)
from ..models.app import AppWallet
from ..models.feature import (
    SendTransactionFeature,
    SignDataFeature,
)
from ..models.wallet import Wallet


def _find_send_transaction_feature(
    features: t.Sequence[t.Any],
) -> SendTransactionFeature | None:
    """Return the first ``SendTransactionFeature`` in *features*, or ``None``."""
    for feature in features:
        if isinstance(feature, SendTransactionFeature):
            return feature
    return None


def _find_sign_data_feature(
    features: t.Sequence[t.Any],
) -> SignDataFeature | None:
    """Return the first ``SignDataFeature`` in *features*, or ``None``."""
    for feature in features:
        if isinstance(feature, SignDataFeature):
            return feature
    return None


def verify_wallet_features(wallet: Wallet) -> None:
    """Verify that the wallet declares required features.

    :param wallet: Connected wallet to validate.
    :raises WalletMissingRequiredFeaturesError: If ``SendTransaction`` is not declared.
    """
    if not wallet.device.features:
        raise WalletMissingRequiredFeaturesError("Wallet did not declare any features")

    has_send_transaction = any(isinstance(f, SendTransactionFeature) for f in wallet.device.features)
    if not has_send_transaction:
        raise WalletMissingRequiredFeaturesError("Wallet does not declare SendTransaction support")


def verify_send_transaction_support(
    wallet: Wallet,
    num_messages: int = 0,
    app_wallet: AppWallet | None = None,
    *,
    has_extra_currency: bool = False,
) -> SendTransactionFeature:
    """Verify ``SendTransaction`` support on both app and wallet layers.

    :param wallet: Connected wallet with device capabilities.
    :param num_messages: Number of outgoing messages (0 to skip check).
    :param app_wallet: Wallet application descriptor, or ``None``.
    :param has_extra_currency: Whether extra-currency transfers are included.
    :return: ``SendTransactionFeature`` from the connected wallet.
    :raises WalletNotSupportFeatureError: If any validation step fails.
    """
    if app_wallet is not None:
        app_feature = _find_send_transaction_feature(app_wallet.features)
        if app_feature is None:
            raise WalletNotSupportFeatureError(
                f"Wallet app `{app_wallet.app_name}` does not declare SendTransaction support"
            )

        if num_messages > 0 and app_feature.max_messages is not None and num_messages > app_feature.max_messages:
            raise WalletNotSupportFeatureError(
                f"Wallet app `{app_wallet.app_name}` supports max "
                f"{app_feature.max_messages} messages, but {num_messages} were requested"
            )

        if has_extra_currency and not app_feature.extra_currency_supported:
            raise WalletNotSupportFeatureError(f"Wallet app `{app_wallet.app_name}` does not support extra currencies")

    wallet_feature = _find_send_transaction_feature(wallet.device.features)
    if wallet_feature is None:
        raise WalletNotSupportFeatureError("Wallet does not support SendTransaction")

    if num_messages > 0 and wallet_feature.max_messages is not None and num_messages > wallet_feature.max_messages:
        raise WalletNotSupportFeatureError(
            f"Wallet supports max {wallet_feature.max_messages} messages, but {num_messages} were requested"
        )

    if has_extra_currency and not wallet_feature.extra_currency_supported:
        raise WalletNotSupportFeatureError("Wallet does not support extra currencies")

    return wallet_feature


def verify_sign_data_support(
    wallet: Wallet,
    payload_type: str | None = None,
    app_wallet: AppWallet | None = None,
) -> SignDataFeature:
    """Verify ``SignData`` support on both app and wallet layers.

    :param wallet: Connected wallet with device capabilities.
    :param payload_type: Payload type to validate (e.g. ``"text"``), or ``None``.
    :param app_wallet: Wallet application descriptor, or ``None``.
    :return: ``SignDataFeature`` from the connected wallet.
    :raises WalletNotSupportFeatureError: If any validation step fails.
    """
    if app_wallet is not None:
        app_feature = _find_sign_data_feature(app_wallet.features)
        if app_feature is None:
            raise WalletNotSupportFeatureError(f"Wallet app `{app_wallet.app_name}` does not declare SignData support")

        if payload_type is not None and payload_type not in app_feature.types:
            raise WalletNotSupportFeatureError(
                f"Wallet app `{app_wallet.app_name}` does not support "
                f"SignData type `{payload_type}`, supported: {app_feature.types}"
            )

    wallet_feature = _find_sign_data_feature(wallet.device.features)
    if wallet_feature is None:
        raise WalletNotSupportFeatureError("Wallet does not support SignData")

    if payload_type is not None and payload_type not in wallet_feature.types:
        raise WalletNotSupportFeatureError(
            f"Wallet does not support SignData type `{payload_type}`, supported: {wallet_feature.types}"
        )

    return wallet_feature


def verify_wallet_network(
    wallet: Wallet,
    expected_network: NetworkGlobalID,
) -> None:
    """Verify the wallet operates on the expected network.

    :param wallet: Connected wallet.
    :param expected_network: Expected network identifier.
    :raises WalletWrongNetworkError: If the network does not match.
    """
    if wallet.account.network != expected_network:
        raise WalletWrongNetworkError(f"Expected network {expected_network.value}, got {wallet.account.network.value}")



from .decoder import EventDecoder
from .message import EventMessage
from .source import EventSource
from .state import ReadyState

__all__ = [
    "EventDecoder",
    "EventMessage",
    "EventSource",
    "ReadyState",
]



import typing as t

from pydantic import (
    BaseModel as _BaseModel,
)
from pydantic import (
    BeforeValidator,
    ConfigDict,
    Field,
    PlainSerializer,
)
from ton_core import Address, Binary, Cell, NetworkGlobalID, PublicKey, StateInit, cell_to_b64, to_cell

__all__ = [
    "A",
    "BaseModel",
    "Binary64",
    "BocCell",
    "ChainId",
    "OptionalBinary64",
    "OptionalBocCell",
    "OptionalChainId",
    "OptionalTonAddress",
    "OptionalTonPublicKey",
    "OptionalWalletStateInit",
    "TonAddress",
    "TonPublicKey",
    "WalletStateInit",
]


T = t.TypeVar("T")
R = t.TypeVar("R")


class BaseModel(_BaseModel):
    """TonConnect base model with alias support and lenient parsing."""

    model_config = ConfigDict(
        populate_by_name=True,
        extra="ignore",
        arbitrary_types_allowed=True,
    )

    def dump(self, **kwargs: t.Any) -> dict[str, t.Any]:
        """Dump to dict using alias keys."""
        return self.model_dump(by_alias=True, **kwargs)

    def dump_json(self, **kwargs: t.Any) -> str:
        """Dump to JSON string using alias keys."""
        return self.model_dump_json(by_alias=True, **kwargs)


# noinspection PyPep8Naming
def A(name: str, default: t.Any = ...) -> t.Any:
    """Create a Pydantic ``Field`` with matching validation and serialization alias.

    :param name: Alias name.
    :param default: Default value, or ``...`` for required.
    :return: Pydantic ``Field``.
    """
    if default is ...:
        return Field(validation_alias=name, serialization_alias=name)
    return Field(default=default, validation_alias=name, serialization_alias=name)


def _conv_opt(conv: t.Callable[[t.Any], T]) -> t.Callable[[t.Any], T | None]:
    """Wrap a converter to pass through ``None``."""

    def _wrap(v: t.Any) -> T | None:
        if v is None:
            return None
        return conv(v)

    return _wrap


def _ser_opt(ser: t.Callable[[T], R]) -> t.Callable[[T | None], R | None]:
    """Wrap a serializer to pass through ``None``."""

    def _wrap(v: T | None) -> R | None:
        if v is None:
            return None
        return ser(v)

    return _wrap


def _as_address(v: t.Any) -> Address:
    return v if isinstance(v, Address) else Address(v)


def _as_network(v: t.Any) -> NetworkGlobalID:
    return v if isinstance(v, NetworkGlobalID) else NetworkGlobalID(int(v))


def _as_public_key(v: t.Any) -> PublicKey:
    return v if isinstance(v, PublicKey) else PublicKey(v)


def _as_state_init(v: t.Any) -> StateInit:
    if isinstance(v, StateInit):
        return v
    return StateInit.deserialize(to_cell(v).begin_parse())


def _as_cell(v: t.Any) -> Cell:
    return v if isinstance(v, Cell) else to_cell(v)


def _as_binary64(v: t.Any) -> Binary:
    return v if isinstance(v, Binary) else Binary(v, size=64)


def _s_address(v: Address) -> str:
    return v.to_str(is_user_friendly=True, is_bounceable=False)


def _s_chain(v: NetworkGlobalID) -> str:
    return str(v.value)


def _s_pubkey(v: PublicKey) -> str:
    return v.as_hex


def _s_state_init(v: StateInit) -> str:
    return cell_to_b64(v.serialize())


def _s_cell(v: Cell) -> str:
    return cell_to_b64(v)


def _s_binary(v: Binary) -> str:
    return v.as_b64


TonAddress: t.TypeAlias = t.Annotated[
    Address,
    BeforeValidator(_as_address),
    PlainSerializer(_s_address, return_type=str),
]

OptionalTonAddress: t.TypeAlias = t.Annotated[
    Address | None,
    BeforeValidator(_conv_opt(_as_address)),
    PlainSerializer(_ser_opt(_s_address), return_type=str | None),
]

ChainId: t.TypeAlias = t.Annotated[
    NetworkGlobalID,
    BeforeValidator(_as_network),
    PlainSerializer(_s_chain, return_type=str),
]

OptionalChainId: t.TypeAlias = t.Annotated[
    NetworkGlobalID | None,
    BeforeValidator(_conv_opt(_as_network)),
    PlainSerializer(_ser_opt(_s_chain), return_type=str | None),
]

TonPublicKey: t.TypeAlias = t.Annotated[
    PublicKey,
    BeforeValidator(_as_public_key),
    PlainSerializer(_s_pubkey, return_type=str),
]

OptionalTonPublicKey: t.TypeAlias = t.Annotated[
    PublicKey | None,
    BeforeValidator(_conv_opt(_as_public_key)),
    PlainSerializer(_ser_opt(_s_pubkey), return_type=str | None),
]

WalletStateInit: t.TypeAlias = t.Annotated[
    StateInit,
    BeforeValidator(_as_state_init),
    PlainSerializer(_s_state_init, return_type=str),
]

OptionalWalletStateInit: t.TypeAlias = t.Annotated[
    StateInit | None,
    BeforeValidator(_conv_opt(_as_state_init)),
    PlainSerializer(_ser_opt(_s_state_init), return_type=str | None),
]

BocCell: t.TypeAlias = t.Annotated[
    Cell,
    BeforeValidator(_as_cell),
    PlainSerializer(_s_cell, return_type=str),
]

OptionalBocCell: t.TypeAlias = t.Annotated[
    Cell | None,
    BeforeValidator(_conv_opt(_as_cell)),
    PlainSerializer(_ser_opt(_s_cell), return_type=str | None),
]

Binary64: t.TypeAlias = t.Annotated[
    Binary,
    BeforeValidator(_as_binary64),
    PlainSerializer(_s_binary, return_type=str),
]

OptionalBinary64: t.TypeAlias = t.Annotated[
    Binary | None,
    BeforeValidator(_conv_opt(_as_binary64)),
    PlainSerializer(_ser_opt(_s_binary), return_type=str | None),
]



import asyncio
import json
import typing as t
from pathlib import Path

from .protocol import StorageProtocol


class FileStorage(StorageProtocol):
    """JSON file-backed key-value storage for TonConnect sessions."""

    def __init__(self, path: Path | str) -> None:
        """Initialize file storage with the given path."""
        self._path = Path(path)
        self._lock = asyncio.Lock()

    async def _read(self) -> dict[str, t.Any]:
        """Read the storage file, returning empty dict if missing."""
        if not self._path.exists():
            return {}

        def _load() -> dict[str, t.Any]:
            with self._path.open("r", encoding="utf-8") as f:
                return t.cast("dict[str, t.Any]", json.load(f))

        return await asyncio.to_thread(_load)

    async def _write(self, data: dict[str, t.Any]) -> None:
        """Write data to the storage file, creating parents as needed."""
        self._path.parent.mkdir(parents=True, exist_ok=True)

        def _dump() -> None:
            with self._path.open("w", encoding="utf-8") as f:
                json.dump(data, f, ensure_ascii=False)

        await asyncio.to_thread(_dump)

    async def set_item(self, key: str, value: t.Any) -> None:
        """Store a value under *key*."""
        async with self._lock:
            data = await self._read()
            data[key] = value
            await self._write(data)

    async def get_item(self, key: str) -> t.Any | None:
        """Return the value for *key*, or ``None``."""
        async with self._lock:
            data = await self._read()
            return data.get(key)

    async def remove_item(self, key: str) -> None:
        """Remove *key* if present."""
        async with self._lock:
            data = await self._read()
            data.pop(key, None)
            await self._write(data)



from __future__ import annotations

import typing as t

from .connector import Connector, Event
from .exceptions import TonConnectError
from .models import (
    ActiveConnection,
    AppWallet,
    SendTransactionResult,
    SignDataResult,
    WalletMessage,
)
from .provider.storage import ProviderStorage

if t.TYPE_CHECKING:
    from .storage import StorageProtocol

_ConnectHandler: t.TypeAlias = t.Callable[
    [Connector, TonConnectError | None],
    t.Awaitable[None],
]
_DisconnectHandler: t.TypeAlias = t.Callable[
    [Connector, TonConnectError | None],
    t.Awaitable[None],
]
_TransactionHandler: t.TypeAlias = t.Callable[
    [Connector, int, SendTransactionResult | None, TonConnectError | None],
    t.Awaitable[None],
]
_SignDataHandler: t.TypeAlias = t.Callable[
    [Connector, int, SignDataResult | None, TonConnectError | None],
    t.Awaitable[None],
]
_MessageHandler: t.TypeAlias = t.Callable[
    [Connector, WalletMessage],
    t.Awaitable[None],
]
_ErrorHandler: t.TypeAlias = t.Callable[
    [Connector, TonConnectError],
    t.Awaitable[None],
]

_EventHandler: t.TypeAlias = (
    _ConnectHandler | _DisconnectHandler | _ErrorHandler | _TransactionHandler | _SignDataHandler | _MessageHandler
)
_EventHandlers: t.TypeAlias = dict[Event, _EventHandler]


class TonConnect:
    """High-level TonConnect manager for multiple connector sessions."""

    def __init__(
        self,
        storage: StorageProtocol,
        manifest_url: str,
        app_wallets: list[AppWallet],
        headers: dict[str, str] | None = None,
        **context: t.Any,
    ) -> None:
        """Initialize the TonConnect manager.

        :param manifest_url: URL to ``tonconnect-manifest.json``.
        :param storage: Key-value storage backend.
        :param app_wallets: Wallet descriptors used as bridge connection sources.
        :param headers: Extra HTTP headers, or ``None``.
        :param context: Arbitrary context values.
        """
        self.storage = storage
        self.manifest_url = manifest_url
        self.app_wallets = app_wallets
        self.headers = headers

        self._handlers: _EventHandlers = {}
        self._context: dict[str, t.Any] = context
        self._connectors: dict[str, Connector] = {}

    def __getitem__(self, key: str) -> t.Any:
        """Return a context value by key."""
        return self._context[key]

    def __setitem__(self, key: str, value: t.Any) -> None:
        """Set a context value by key."""
        self._context[key] = value

    @property
    def context(self) -> dict[str, t.Any]:
        """User context dict."""
        return self._context

    @t.overload
    def register(
        self,
        event: t.Literal[Event.CONNECT],
        handler: _ConnectHandler,
    ) -> _ConnectHandler: ...

    @t.overload
    def register(
        self,
        event: t.Literal[Event.DISCONNECT],
        handler: _DisconnectHandler,
    ) -> _DisconnectHandler: ...

    @t.overload
    def register(
        self,
        event: t.Literal[Event.TRANSACTION],
        handler: _TransactionHandler,
    ) -> _TransactionHandler: ...

    @t.overload
    def register(
        self,
        event: t.Literal[Event.SIGN_DATA],
        handler: _SignDataHandler,
    ) -> _SignDataHandler: ...

    @t.overload
    def register(
        self,
        event: t.Literal[Event.MESSAGE],
        handler: _MessageHandler,
    ) -> _MessageHandler: ...

    @t.overload
    def register(
        self,
        event: t.Literal[Event.ERROR],
        handler: _ErrorHandler,
    ) -> _ErrorHandler: ...

    def register(
        self,
        event: Event,
        handler: _EventHandler,
    ) -> _EventHandler:
        """Register an event handler.

        :param event: Event type.
        :param handler: Async handler callable.
        :return: The registered handler.
        """
        self._handlers[event] = handler
        return handler

    @t.overload
    def on(
        self,
        event: t.Literal[Event.CONNECT],
    ) -> t.Callable[[_ConnectHandler], _ConnectHandler]: ...

    @t.overload
    def on(
        self,
        event: t.Literal[Event.DISCONNECT],
    ) -> t.Callable[[_DisconnectHandler], _DisconnectHandler]: ...

    @t.overload
    def on(
        self,
        event: t.Literal[Event.TRANSACTION],
    ) -> t.Callable[[_TransactionHandler], _TransactionHandler]: ...

    @t.overload
    def on(
        self,
        event: t.Literal[Event.SIGN_DATA],
    ) -> t.Callable[[_SignDataHandler], _SignDataHandler]: ...

    @t.overload
    def on(
        self,
        event: t.Literal[Event.MESSAGE],
    ) -> t.Callable[[_MessageHandler], _MessageHandler]: ...

    @t.overload
    def on(
        self,
        event: t.Literal[Event.ERROR],
    ) -> t.Callable[[_ErrorHandler], _ErrorHandler]: ...

    def on(
        self,
        event: Event,
    ) -> t.Callable[..., t.Any]:
        """Register an event handler via decorator.

        :param event: Event type.
        """

        def decorator(handler: _EventHandler) -> _EventHandler:
            self._handlers[event] = handler
            return handler

        return decorator

    @property
    def connectors(self) -> dict[str, Connector]:
        """Active connectors by session key (read-only copy)."""
        return dict(self._connectors)

    def create_connector(
        self,
        session_key: str,
        **context: t.Any,
    ) -> Connector:
        """Create a new ``Connector`` for a session.

        :param session_key: Unique session key.
        :param context: Extra context values merged with the global context.
        :return: New connector instance.
        """
        provider_storage = ProviderStorage(
            storage=self.storage,
            session_key=session_key,
        )
        connector = Connector(
            storage=provider_storage,
            session_key=session_key,
            manifest_url=self.manifest_url,
            app_wallets=self.app_wallets,
            handlers=dict(self._handlers),
            headers=self.headers,
            context={**self._context, **context},
        )
        self._connectors[session_key] = connector
        return connector

    async def from_connection(
        self,
        connection: ActiveConnection,
        session_key: str,
        **context: t.Any,
    ) -> Connector:
        """Reconstruct a connector from a saved ``ActiveConnection``.

        :param connection: Stored active connection.
        :param session_key: Session key for the connector.
        :param context: Extra context values.
        :return: Restored connector.
        """
        connector = self.create_connector(session_key, **context)
        await connector.storage.store_connection(connection)
        await connector.restore()
        return connector

    async def close_all(self) -> None:
        """Close all active connectors."""
        for connector in self._connectors.values():
            await connector.close()
        self._connectors.clear()



import time
import typing as t

from ._types import A, BaseModel
from .request import ConnectRequest
from .response import ConnectEventSuccess
from .session import BridgeProviderSession, SessionKeyPair


class ConnectionSource(BaseModel):
    """Bridge endpoint with universal link."""

    bridge_url: str = A("bridgeUrl")
    """Bridge base URL."""
    universal_link: str = A("universalLink")
    """Wallet universal link."""


ConnectionSources = list[ConnectionSource]


class PendingConnection(BaseModel):
    """Pending (not yet accepted) TonConnect connection."""

    connect_request: ConnectRequest = A("connectRequest")
    """Original connect request."""
    connection_sources: ConnectionSources = A("connectionSources")
    """Available bridge endpoints."""
    session_keypair: SessionKeyPair = A("sessionKeyPair")
    """Session X25519 key pair."""
    created_at: int | None = A("createdAt", default=None)
    """Creation unix timestamp, or ``None``."""

    def is_expired(self) -> bool:
        """Check whether the pending connection has expired (15 min TTL)."""
        if self.created_at is not None:
            now = int(time.time())
            return (now - self.created_at) > 15 * 60
        return True


class ActiveConnection(BaseModel):
    """Established TonConnect connection."""

    type: t.Literal["http"] = "http"
    """Connection type."""
    connect_event: ConnectEventSuccess = A("connectEvent")
    """Successful connect event."""
    session: BridgeProviderSession
    """Bridge provider session."""
    next_rpc_request_id: int = A("nextRpcRequestId", default=0)
    """Next RPC request sequence number."""
    last_wallet_event_id: int | str | None = A("lastWalletEventId", default=None)
    """Last wallet event ID, or ``None``."""
    last_event_id: str | None = A("lastEventId", default=None)
    """Last SSE event ID, or ``None``."""


Connection: t.TypeAlias = ActiveConnection | PendingConnection



from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True)
class EventMessage:
    """Parsed SSE event message."""

    event: str = "message"
    """Event type name."""
    data: str | None = None
    """Event payload, or ``None``."""
    event_id: str | None = None
    """Event identifier, or ``None``."""

    @classmethod
    def parse(cls, raw: str) -> EventMessage:
        """Parse a raw SSE event block into an ``EventMessage``.

        :param raw: Raw SSE text block (lines separated by newlines).
        :return: Parsed event message.
        """
        event: str = "message"
        data_parts: list[str] = []
        event_id: str | None = None

        for line in raw.splitlines():
            if not line or line.startswith(":"):
                continue

            field, sep, value = line.partition(":")
            if not sep:
                continue

            if value.startswith(" "):
                value = value[1:]

            if field == "data":
                data_parts.append(value)
            elif field == "event":
                event = value or "message"
            elif field == "id":
                event_id = value or None

        data = "\n".join(data_parts) if data_parts else None
        return cls(event=event, data=data, event_id=event_id)



from .connector import Connector, Event
from .storage import (
    FileStorage,
    MemoryStorage,
    StorageProtocol,
)
from .tonconnect import TonConnect
from .utils import (
    AppWalletsLoader,
    VerifySignData,
    VerifyTonProof,
    create_ton_proof_payload,
    verify_ton_proof_payload,
)

__all__ = [
    "AppWalletsLoader",
    "Connector",
    "Event",
    "FileStorage",
    "MemoryStorage",
    "StorageProtocol",
    "TonConnect",
    "VerifySignData",
    "VerifyTonProof",
    "create_ton_proof_payload",
    "verify_ton_proof_payload",
]



import typing as t

__all__ = [
    "BadRequestError",
    "FetchWalletsError",
    "ManifestContentError",
    "ManifestNotFoundError",
    "MethodNotSupportedError",
    "RequestTimeoutError",
    "TonConnectError",
    "TonConnectErrors",
    "UnknownAppError",
    "UnknownError",
    "UserRejectsError",
    "WalletAlreadyConnectedError",
    "WalletMissingRequiredFeaturesError",
    "WalletNotConnectedError",
    "WalletNotSupportFeatureError",
    "WalletWrongNetworkError",
]


class TonConnectError(Exception):
    """Base exception for all TonConnect errors."""

    default_message: t.ClassVar[str] = "TonConnect error."
    """Default error description used when no message is provided."""
    default_info: t.ClassVar[str | None] = None
    """Default extended explanation, or ``None``."""

    code: int
    """Numeric RPC error code."""
    message: str
    """Error description."""
    info: str | None
    """Extended explanation, or ``None``."""

    def __init__(
        self,
        message: str | None = None,
        *,
        code: int = 0,
        info: str | None = None,
    ) -> None:
        """Initialize the error.

        :param message: Error description, or ``None`` for ``default_message``.
        :param code: Numeric RPC error code.
        :param info: Extended explanation, or ``None`` for ``default_info``.
        """
        self.code = int(code)
        self.message = message or self.default_message
        self.info = self.default_info if info is None else info

        text = self.message
        if self.info:
            text = f"{text} ({self.info})"
        super().__init__(text)


class WalletAlreadyConnectedError(TonConnectError):
    """Wallet is already connected."""

    default_message = "Wallet already connected."
    default_info = (
        "Wallet connection called but wallet already connected. "
        "To avoid the error, disconnect the wallet before doing a new connection."
    )


class WalletNotConnectedError(TonConnectError):
    """No wallet connected for the requested RPC method."""

    default_message = "Wallet not connected."
    default_info = "Send transaction or other protocol methods called while wallet is not connected."


class WalletMissingRequiredFeaturesError(TonConnectError):
    """Wallet does not declare minimum required features."""

    default_message = "Missing required features."
    default_info = "Update the wallet application."


class WalletNotSupportFeatureError(TonConnectError):
    """Wallet does not support the requested feature or capability."""

    default_message = "Wallet doesn't support the requested feature method."
    default_info = "Wallet doesn't support the requested feature method."


class WalletWrongNetworkError(TonConnectError):
    """Wallet operates on a different network than expected."""

    default_message = "Wallet connected to wrong network."
    default_info = "Switch the wallet network and try again."


class FetchWalletsError(TonConnectError):
    """Wallets list could not be fetched or parsed."""

    default_message = "Failed to fetch wallets list."
    default_info = "An error occurred while fetching the wallets list."


class UnknownError(TonConnectError):
    """Unrecognised TonConnect RPC error (code 0)."""

    default_message = "Unknown error occurred."
    default_info = "Unknown error."


class BadRequestError(TonConnectError):
    """Malformed RPC request (code 1)."""

    default_message = "Bad request."
    default_info = "Request to the wallet contains errors."


class UnknownAppError(TonConnectError):
    """RPC request from an unconnected app (code 100)."""

    default_message = "App RPC request error."
    default_info = "App tries to send RPC request to the injected wallet while not connected."


class UserRejectsError(TonConnectError):
    """User rejected the action in the wallet UI (code 300)."""

    default_message = "User rejected the action."
    default_info = "User rejects the action in the wallet."


class ManifestNotFoundError(TonConnectError):
    """Wallet could not fetch ``tonconnect-manifest.json`` (code 2)."""

    default_message = "Manifest not found."
    default_info = (
        "Make sure you added tonconnect-manifest.json to the root of your app "
        "or passed correct manifest_url. See more at "
        "https://github.com/ton-connect/docs/blob/main/requests-responses.md#app-manifest"
    )


class ManifestContentError(TonConnectError):
    """``tonconnect-manifest.json`` content is invalid (code 3)."""

    default_message = "Manifest content error."
    default_info = (
        "The passed tonconnect-manifest.json contains errors. "
        "Check the format of your manifest. See more at "
        "https://github.com/ton-connect/docs/blob/main/requests-responses.md#app-manifest"
    )


class MethodNotSupportedError(TonConnectError):
    """Wallet does not support the requested RPC method (code 400)."""

    default_message = "Method not supported."
    default_info = "The requested method is not supported by the wallet or is unknown."


class RequestTimeoutError(TonConnectError):
    """User did not respond within the time limit (code 500)."""

    default_message = "Request timed out."
    default_info = "User did not respond within the predefined time limit."


class TonConnectErrors:
    """Map TonConnect RPC error codes to typed exceptions."""

    CODE_TO_ERROR: t.ClassVar[dict[int, type[TonConnectError]]] = {
        0: UnknownError,
        1: BadRequestError,
        2: ManifestNotFoundError,
        3: ManifestContentError,
        100: UnknownAppError,
        300: UserRejectsError,
        400: MethodNotSupportedError,
        500: RequestTimeoutError,
    }
    """Mapping of RPC error codes to exception types."""

    @classmethod
    def from_code(cls, code: int, message: str) -> TonConnectError:
        """Create a typed ``TonConnectError`` from an RPC error code.

        :param code: Numeric error code from the wallet.
        :param message: Error message from the wallet.
        :return: Corresponding exception instance.
        """
        exc_type = cls.CODE_TO_ERROR.get(int(code), TonConnectError)
        return exc_type(message, code=int(code))



from __future__ import annotations

from ..exceptions import TonConnectError
from ._types import A, BaseModel
from .account import Account
from .device import Device
from .dto import TonProofPayloadDto
from .proof import TonProofData
from .response import (
    ConnectEventPayload,
    TonAddressItemReply,
    TonProofItemReply,
)


class Wallet(BaseModel):
    """Connected wallet state."""

    device: Device
    """Wallet device information."""
    account: Account
    """Wallet account data."""
    ton_proof: TonProofData | None = A("tonProof", default=None)
    """TON Proof data, or ``None``."""

    @property
    def ton_proof_dto(self) -> TonProofPayloadDto | None:
        """Build a ``TonProofPayloadDto`` from the stored proof, or ``None``."""
        if self.ton_proof is None:
            return None
        return TonProofPayloadDto(
            address=self.account.address,
            network=self.account.network,
            public_key=self.account.public_key,
            wallet_state_init=self.account.state_init,
            proof=self.ton_proof,
        )

    @classmethod
    def from_payload(cls, payload: ConnectEventPayload) -> Wallet:
        """Create a ``Wallet`` from a ``ConnectEventPayload``.

        :param payload: Successful connect event payload.
        :return: Wallet instance.
        :raises TonConnectError: If payload lacks a ``ton_addr`` item.
        """
        account: Account | None = None
        ton_proof: TonProofData | None = None

        for item in payload.items:
            if isinstance(item, TonAddressItemReply):
                account = Account.model_validate(item.model_dump(by_alias=True))
            elif isinstance(item, TonProofItemReply):
                ton_proof = item.proof

        if account is None:
            raise TonConnectError("ConnectEventPayload does not contain required `ton_addr` item")

        return cls(device=payload.device, account=account, ton_proof=ton_proof)



import typing as t

from pydantic import Field

from ._types import A, BaseModel


class SendTransactionFeature(BaseModel):
    """Wallet ``SendTransaction`` feature declaration."""

    name: t.Literal["SendTransaction"]
    """Feature name literal."""
    max_messages: int | None = A("maxMessages", default=None)
    """Maximum outgoing messages, or ``None``."""
    extra_currency_supported: bool = A("extraCurrencySupported", default=False)
    """Whether extra currencies are supported."""


class SignDataFeature(BaseModel):
    """Wallet ``SignData`` feature declaration."""

    name: t.Literal["SignData"]
    """Feature name literal."""
    types: list[t.Literal["binary", "cell", "text"]]
    """Supported payload types."""


FeatureType: t.TypeAlias = t.Annotated[
    SendTransactionFeature | SignDataFeature,
    Field(discriminator="name"),
]
FeatureTypes = list[FeatureType]



import asyncio
import typing as t
from contextlib import suppress

from pydantic import TypeAdapter

from ..exceptions import TonConnectError
from ..models import (
    ActiveConnection,
    AppWallet,
    BridgeProviderSession,
    ConnectEventSuccess,
    ConnectionSource,
    ConnectRequest,
    DisconnectEventSuccess,
    DisconnectRpcRequest,
    EventBase,
    IncomingMessage,
    PendingConnection,
    RpcRequestBase,
    SessionKeyPair,
    WalletMessage,
    WalletResponse,
)
from .gateway import Gateway, retry_until_success
from .storage import ProviderStorage

_WALLET_MESSAGE_ADAPTER: TypeAdapter[WalletMessage] = TypeAdapter(WalletMessage)


class Provider:
    """Bridge-level provider managing SSE gateways and message routing."""

    SEND_TTL: int = 300
    SEND_ATTEMPTS: int = 10
    SEND_DELAY: float = 5.0

    CONNECT_ATTEMPTS: int = 10
    CONNECT_TIMEOUT: float = 12.0
    RECONNECT_DELAY: float = 2.0

    def __init__(
        self,
        *,
        storage: ProviderStorage,
        app_wallets: list[AppWallet],
        headers: dict[str, str] | None = None,
        on_message: t.Callable[[WalletMessage], t.Awaitable[None]],
        on_error: t.Callable[[Exception], t.Awaitable[None]],
    ) -> None:
        """Initialize the provider.

        :param storage: Provider storage.
        :param app_wallets: Wallet descriptors from which bridge connection sources are derived.
        :param headers: Extra HTTP headers, or ``None``.
        :param on_message: Async callback for wallet messages.
        :param on_error: Async callback for errors.
        """
        self._storage = storage
        self._connection_sources = self._to_connection_sources(app_wallets)

        self._session_keypair = SessionKeyPair.generate()
        self._session_id = self._session_keypair.session_id

        self._headers = headers or {}
        self._on_message = on_message
        self._on_error = on_error

        self._gateway: Gateway | None = None
        self._pending_gateways: list[Gateway] = []

        self._connect_lock = asyncio.Lock()

    @property
    def session_id(self) -> str:
        """Hex-encoded public key used as the SSE client ID."""
        return self._session_id

    @property
    def storage(self) -> ProviderStorage:
        """Provider storage instance."""
        return self._storage

    @property
    def bridge_url(self) -> str | None:
        """Bridge URL of the active gateway, or ``None``."""
        return self._gateway.bridge_url if self._gateway else None

    async def connect(self, request: ConnectRequest) -> None:
        """Store a pending connection and open SSE gateways.

        :param request: Connect request payload.
        """
        pending = PendingConnection(
            session_keypair=self._session_keypair,
            connect_request=request,
            connection_sources=self._connection_sources,
        )
        await self._storage.store_connection(pending)
        await self._open_gateways()

    async def request(self, request: RpcRequestBase) -> int:
        """Encrypt and send an RPC request to the wallet.

        :param request: RPC request to send.
        :return: Assigned request ID.
        :raises TonConnectError: If no active connection or gateway.
        """
        conn = await self._storage.get_connection()
        if not isinstance(conn, ActiveConnection):
            raise TonConnectError("Wallet is not connected")

        gw = self._gateway
        if gw is None:
            raise TonConnectError("Bridge gateway is not initialized")

        request_id = await self._storage.get_next_rpc_request_id()
        request.id = str(request_id)

        message = conn.session.session_keypair.encrypt(
            message=request.to_bytes(),
            receiver_public_key=conn.session.wallet_public_key,
        )
        await gw.send(
            message,
            conn.session.receiver,
            request.method,
            ttl=self.SEND_TTL,
            attempts=self.SEND_ATTEMPTS,
            delay=self.SEND_DELAY,
        )

        await self._storage.increase_next_rpc_request_id()
        return request_id

    async def disconnect(self) -> None:
        """Send a disconnect RPC (best-effort) and tear down the connection."""
        try:
            conn = await self._storage.get_connection()
        except TonConnectError:
            conn = None

        if isinstance(conn, ActiveConnection) and self._gateway is not None:
            request = DisconnectRpcRequest()

            with suppress(Exception):
                request_id = await self._storage.get_next_rpc_request_id()
                request.id = str(request_id)

                message = conn.session.session_keypair.encrypt(
                    message=request.to_bytes(),
                    receiver_public_key=conn.session.wallet_public_key,
                )
                await self._gateway.send(
                    message,
                    conn.session.receiver,
                    request.method,
                    ttl=self.SEND_TTL,
                    attempts=self.SEND_ATTEMPTS,
                    delay=self.SEND_DELAY,
                )

        await self._storage.remove_connection()
        await self.close_connection()

    async def restore_connection(self) -> ConnectEventSuccess | None:
        """Restore a previously saved connection from storage.

        :return: Stored connect event on success, or ``None``.
        """
        try:
            conn = await self._storage.get_connection()
        except TonConnectError:
            return None

        if isinstance(conn, PendingConnection):
            self._session_keypair = conn.session_keypair
            self._session_id = self._session_keypair.session_id
            await self._close_gateways()
            await self._open_gateways()
            return None

        if not isinstance(conn, ActiveConnection):
            return None  # type: ignore[unreachable]

        self._session_keypair = conn.session.session_keypair
        self._session_id = self._session_keypair.session_id

        await self._close_gateways()

        self._gateway = Gateway(
            storage=self._storage,
            session_id=self._session_id,
            bridge_url=conn.session.bridge_url,
            headers=self._headers,
            on_gateway_message=self._on_gateway_message,
            on_gateway_error=self._on_gateway_error,
        )

        try:
            await retry_until_success(
                self._gateway.register_session,
                attempts=self.CONNECT_ATTEMPTS,
                delay=self.RECONNECT_DELAY,
            )
        except Exception:
            await self.close_connection()
            return None

        return conn.connect_event

    async def close_connection(self) -> None:
        """Close the active gateway and all pending gateways."""
        gw = self._gateway
        pending = list(self._pending_gateways)

        self._gateway = None
        self._pending_gateways = []

        if gw is not None:
            await gw.close()

        for pgw in pending:
            await pgw.close()

    async def pause(self) -> None:
        """Pause SSE listening on all gateways."""
        if self._gateway is not None:
            await self._gateway.pause()
        for gw in self._pending_gateways:
            await gw.pause()

    async def unpause(self) -> None:
        """Resume SSE listening on all gateways."""
        if self._gateway is not None:
            await self._gateway.unpause()
        for gw in self._pending_gateways:
            await gw.unpause()

    @staticmethod
    def _to_connection_sources(
        wallets: list[AppWallet],
    ) -> list[ConnectionSource]:
        """Convert wallet descriptors to connection sources."""
        return [
            ConnectionSource(
                bridge_url=wallet.bridge_url,
                universal_link=wallet.universal_url or "",
            )
            for wallet in wallets
            if wallet.bridge_url is not None
        ]

    async def _open_gateways(self) -> None:
        """Open SSE gateways for all connection sources."""
        await self.close_connection()

        for connection_source in self._connection_sources:
            gw = Gateway(
                storage=self._storage,
                session_id=self._session_id,
                bridge_url=connection_source.bridge_url,
                headers=self._headers,
                on_gateway_message=self._on_gateway_message,
                on_gateway_error=self._on_gateway_error,
            )
            self._pending_gateways.append(gw)

        if self._pending_gateways:
            tasks = [gw.register_session(timeout=2) for gw in self._pending_gateways]
            results = await asyncio.gather(*tasks, return_exceptions=True)

            alive = []
            first_error: BaseException | None = None
            for gw, result in zip(self._pending_gateways, results, strict=True):
                if isinstance(result, BaseException):
                    if first_error is None:
                        first_error = result
                    await gw.close()
                else:
                    alive.append(gw)

            self._pending_gateways = alive

            if not self._pending_gateways and first_error is not None:
                raise first_error

    async def _close_gateways(self, *, exclude: Gateway | None = None) -> None:
        """Close all pending gateways, optionally keeping one."""
        pending = list(self._pending_gateways)
        self._pending_gateways = []

        for pgw in pending:
            if pgw is not exclude:
                await pgw.close()

    async def _select_gateway(self, bridge_url: str) -> None:
        """Promote a pending gateway to active by bridge URL."""
        keep: Gateway | None = None
        for gw in self._pending_gateways:
            if gw.bridge_url == bridge_url:
                keep = gw
                break

        if keep is None:
            return

        await self._close_gateways(exclude=keep)
        self._gateway = keep
        self._pending_gateways = []

    async def _on_gateway_error(self, exc: Exception) -> None:
        """Forward gateway errors to the error callback."""
        await self._on_error(exc)

    async def _on_gateway_message(
        self,
        msg: IncomingMessage,
        bridge_url: str,
    ) -> None:
        """Route an incoming gateway message to the handler."""
        try:
            await self._handle_message(msg, bridge_url)
        except Exception as exc:
            await self._on_error(exc)

    async def _handle_message(
        self,
        msg: IncomingMessage,
        bridge_url: str,
    ) -> None:
        """Decrypt, parse, and dispatch an incoming bridge message."""
        try:
            conn = await self._storage.get_connection()
        except TonConnectError:
            return

        if isinstance(conn, PendingConnection):
            session_keypair = conn.session_keypair
        elif isinstance(conn, ActiveConnection):
            session_keypair = conn.session.session_keypair
        else:
            return  # type: ignore[unreachable]

        decrypted = session_keypair.decrypt(msg.message, msg.sender_public_key)

        wallet_msg = _WALLET_MESSAGE_ADAPTER.validate_json(decrypted)
        if isinstance(wallet_msg, WalletResponse):
            await self._on_message(wallet_msg)
            return

        # Deduplicate wallet events by monotonic ID.
        if isinstance(wallet_msg, EventBase) and wallet_msg.id is not None:
            last_id = await self._storage.get_last_wallet_event_id()
            if (
                last_id is not None and type(wallet_msg.id) is type(last_id) and wallet_msg.id <= last_id  # type: ignore[operator]
            ):
                return

        # Lock to prevent race when multiple bridges deliver
        # ConnectEventSuccess concurrently.
        if isinstance(conn, PendingConnection) and isinstance(wallet_msg, ConnectEventSuccess):
            async with self._connect_lock:
                try:
                    inner_conn = await self._storage.get_connection()
                except TonConnectError:
                    inner_conn = None

                if isinstance(inner_conn, PendingConnection):
                    active = ActiveConnection(
                        connect_event=wallet_msg,
                        next_rpc_request_id=0,
                        last_wallet_event_id=wallet_msg.id,
                        session=BridgeProviderSession(
                            session_keypair=session_keypair,
                            wallet_public_key=msg.sender_public_key,
                            bridge_url=bridge_url,
                        ),
                    )
                    await self._storage.store_connection(active)
                    await self._select_gateway(bridge_url)
                else:
                    return
        elif isinstance(wallet_msg, EventBase) and wallet_msg.id is not None:
            await self._storage.store_last_wallet_event_id(wallet_msg.id)

        if isinstance(wallet_msg, DisconnectEventSuccess):
            await self._storage.remove_connection()
            await self.close_connection()

        await self._on_message(wallet_msg)
